;################################################################################
;#										#
;# libmio - multi i/o for ATMega644 version 0.10				#
;# copyright (c) 2005-2008 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;# * PAL/NTSC RGB videodisplay with 30x23 chars and 8 colors			#
;# * 1 channel sound can play notes and noise					#
;# * ps2-keyboard interface							#
;# * 1200 Bps bidirectional serial port						#
;# * parallel printer port							#
;# * pseudorandom generator							#
;# * 0,1s timer									#
;#										#
;# This library is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU Lesser General Public			#
;# License as published by the Free Software Foundation; either			#
;# version 2 of the License, or (at your option) any later version.		#
;#										#
;# This library is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU		#
;# Lesser General Public License for more details.				#
;#										#
;# You should have received a copy of the GNU Lesser General Public		#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################
;
;Interface:
;
;	libmio_isr	: video interrupt (compare match OCR1B)
;	libmio_init	: initialize timer&port 
; 	libmio_outd	: output dezimal value (XL,XH) with format (tempreg2)
; 	libmio_outh	: output hexadezimal value (XL) with format (tempreg2)
; 	libmio_clr	: clear screen
; 	libmio_fill	: fill screen with char tempreg1
; 	libmio_orom	: output null-teminated string from flash (Z)
; 	libmio_oram	: output null-teminated string from RAM (Z)
; 	libmio_mtext	: output multiple chars (tempreg1, ZL=n)
; 	libmio_char	: output character (tempreg1)
; 	libmio_newline	: output newline (goto start of next line)
; 	libmio_space	: output space character
; 	libmio_comma	: output comma character
;	libmio_plot	: plot/unplot pixel
;	libmio_box	: draw rectangle
;	libmio_fbox	: draw filled rectangle
;	libmio_draw	: draw a line
;	libmio_note	: play a note or noise
;	libmio_waitkey	: wait until a key is pressed
;	libmio_cursor	: wait until a key is pressed and blink Cursor
;	libmio_pser	: put tempreg1 to serial output
;	libmio_pser	: get tempreg1 from serial input
;	libmio_ppar	: put tempreg1 to parallel output
;	libmio_screenshot send screenshot to serial output
;	libmio_chp	: enable/disable charge pump (XL)
;	libmio_cconv	: convert character output
;	libmio_pconv	: convert character input
;
;###############################################################################
.set use_libmio_editor_mode = 1

.ifdef use_libmio_all
    .set	use_libmio_outd = 1
    .set	use_libmio_outh = 1
    .set	use_libmio_orom = 1
    .set	use_libmio_oram = 1
    .set	use_libmio_othis = 1
    .set	use_libmio_mtext = 1
    .set	use_libmio_newline = 1
    .set	use_libmio_gotoxy = 1
    .set	use_libmio_plot = 1
    .set	use_libmio_box = 1
    .set	use_libmio_fbox = 1
    .ser	use_libmio_draw = 1
    .set	use_libmio_bigchar = 1
    .set	use_libmio_sync = 1
    .set	use_libmio_note = 1
    .set	use_libmio_waitkey = 1
    .set	use_libmio_cursor = 1
    .set	use_libmio_pser = 1
    .set	use_libmio_gser = 1
    .set	use_libmio_ppar = 1
    .set	use_libmio_screenshot = 1
    .set	use_libmio_chp = 1
    .set	use_libmio_cconv = 1
    .set	use_libmio_pconv = 1
.endif


libmio_isr_im:	rjmp	libmio_iss		;jump extender

;-------------------------------------------------------------------------------		
;video-interrupt (no registers destroyed, stack=13)
;save registers
;-------------------------------------------------------------------------------		
libmio_isr:	push	tempreg1		;2
		push	tempreg2		;2
		push	tempreg3		;2
		push	ZH			;2 save Z-register
		push	ZL			;2 save Z-register
		push	YH			;2 save Y-register
		push	YL			;2 save Y-register
		push	XH			;2 save Y-register
		push	XL			;2 save Y-register
		push	r0			;2 save r0
		push	r1			;2 save r1		
		in	ZH,SREG			;1 save status
		push	ZH			;2

;-------------------------------------------------------------------------------		
;sound output
;-------------------------------------------------------------------------------
		lds	XH,libmio_pwm		;2 pwm value
		sts	OCR2A,XH		;1 set level
;+40c
;-------------------------------------------------------------------------------		
;serial port io
;-------------------------------------------------------------------------------
		lds	XH,libmio_serio		;2 serial io
		sbrs	XH,7			;1 skip if 1-bit
		cbi	PORTD,4			;1 clear pin
		sbrc	XH,7			;1 skip if 0-bit
		sbi	PORTD,4			;1 set pin
		sbic	PIND,3			;1 check signal
		andi	XH,0xfe			;1
		sbis	PIND,3			;1 check signal
		ori	XH,0x01			;1
		sts	libmio_serio,XH		;2
		
;-------------------------------------------------------------------------------		
;count line up
;-------------------------------------------------------------------------------
		add	vline_l,byte1		;1
		adc	vline_h,byte0		;1

;-------------------------------------------------------------------------------		
;check if line is image
;-------------------------------------------------------------------------------
		lds	ZL,libmio_config	;2 configuration
		ldi	ZH,230			;1 number of clines
		lds	XL,libmio_vidmode	;2 video mode
		cpi	XL,0x00
		breq	libmio_ismode0
		ldi	ZH,232			;1 number of clines in VM1+3
		cpi	XL,0x02
		brne	libmio_ismode1
		ldi	ZH,228
		rjmp	libmio_ismode1
		
libmio_ismode0:	sbrc	ZL,7			;1 skip if normal mode
		ldi	ZH,228			;1 number of clines in editor mode
libmio_ismode1:	cp	vline_l,ZH		;1 max lines
		sbrs	vline_h,0		;1 skip if no image
		brcs	libmio_isr_im		;2 goto image-line

;-------------------------------------------------------------------------------
;no image line
;-------------------------------------------------------------------------------		
libmio_is_02:	brne	libmio_is_08		;2 no maxline
		lds	YL,libmio_vflags	;odd/even
		sbrc	YL,7
		add	vline_l,byte1		;1 add 1 line
		subi	YL,0x80			;1 swap odd/even flag
		ldi	XL,5			;1 fps/10 for pal
		sbrs	ZL,2			;1 pal/ntsc bit
		inc	XL			;1 fps/10 for ntsc
		ldi	tempreg1,50		;1 fps for pal
		sbrs	ZL,2			;1 pal/ntsc bit
		ldi	tempreg1,60		;1 fps for ntsc
		
		lds	XH,libmio_time0		;2 timer
		inc	XH
		cp	XH,XL			;1 frames (5/6)
		brne	libmio_is_06a
		clr	XH
		lds	ZL,libmio_time_l
		lds	ZH,libmio_time_h
		adiw	ZL,1
		sts	libmio_time_l,ZL
		sts	libmio_time_h,ZH		
libmio_is_06a:	sts	libmio_time0,XH		
		lds	YH,libmio_frames	;2 get frame counter
		inc	YH			;1 +1
		cp	YH,tempreg1		;1 frames (50/60)
		brne	libmio_is_07		;2 frame oveflow
		clr	YH			;1 set frames=0
		ori	YL,0x40			;1 set seconds flag
libmio_is_07:	sts	libmio_frames,YH	;2 set frames=0
		sts	libmio_vflags,YL		
		rjmp	libmio_is_15		;2
		
libmio_is_08:	ldi	ZH,libmio_vstart_p	;1 vstart pal
		sbrs	ZL,2			;1 pal/ntsc bit
		ldi	ZH,libmio_vstart_n	;1 vstart ntsc
		cp	vline_l,ZH
		brne	libmio_is_09		;no vsync start
		cbi	PORTD,6
		sbrs	ZL,3			;csync bit
		rjmp	libmio_is_15  
		ldi	XH,0x82			;invert CSYNC
		sts	TCCR1A,XH
		rjmp	libmio_is_15  


libmio_is_09:	ldi	ZH,libmio_vstop_p	;1 vstop pal
		sbrs	ZL,2			;1 pal/ntsc bit
		ldi	ZH,libmio_vstop_n	;1 vstop ntsc
		cp	vline_l,ZH
		brne	libmio_is_10		;no vsync stop
		sbi	PORTD,6
		sbrs	ZL,3			;csync bit
		rjmp	libmio_is_15  
		ldi	XH,0xc2			;normal CSYNC
		sts	TCCR1A,XH
		rjmp	libmio_is_15  

		
libmio_is_10:	ldi	ZH,57			;1 maxlines pal
		sbrs	ZL,2			;1 pal/ntsc bit
		ldi	ZH,7			;1 maxlines ntsc
		cp	vline_l,ZH
		breq	libmio_is_11		;2 end of frame		
		rjmp	libmio_is_15  
		
libmio_is_11:	ldi	XL,LOW(libmio_vram)	;1 set vram start
		sts	libmio_rampos_l,XL	;2
		ldi	XL,HIGH(libmio_vram)	;1
		sts	libmio_rampos_h,XL	;2
		sts	libmio_cline,byte0	;2 store cline
		clr	vline_l			;line_l=0
		clr	vline_h			;line_h=0
		ldi	XL,0x0a			;editor head color
		sts	libmio_elcolor,XL

;-------------------------------------------------------------------------------
; calculate envelope position
;-------------------------------------------------------------------------------		
		ldi	ZH,HIGH(libmio_etable*2);1 start envtable high
		ldi	ZL,LOW(libmio_etable*2)	;1 start envtable low
		lds	XL,libmio_env		;2 envelope pointer
		add	ZL,XL			;1 envelope-position
		adc	ZH,byte0		;1
		lpm	XH,Z			;get envelope value
		cp	XH,byte0		;must be zero to stop
		breq	libmio_is_11a		;the end is reached
		inc	XL			;+1
		sts	libmio_env,XL		;write back
libmio_is_11a:	sts	libmio_envv,XH		;write envelope value
		rjmp	libmio_is_15

;-------------------------------------------------------------------------------
;sync to counter
;-------------------------------------------------------------------------------		
libmio_iss:	lds	XL,TCNT1L		;1 get timer
		cpi	XL,libmio_hstart+78 	;1
		brcs	libmio_iss		;1 loop		
libmio_iss_01:	cpi	XL,libmio_hstart+79	
		brcs	libmio_iss_02
libmio_iss_02:	cpi	XL,libmio_hstart+80	
		brcs	libmio_iss_03
libmio_iss_03:	cpi	XL,libmio_hstart+81	
		brcs	libmio_iss_04
libmio_iss_04:	sbrc	ZL,5			;1 fast flag
		rjmp	libmio_is_15		;2 no visible lines

		lds	ZL,libmio_config
		sbrc	ZL,7
		jmp	libmio_tvm		;editor mode
		lds	ZL,libmio_vidmode
		andi	ZL,0x03			;only 4 modes
		brne	libmio_iss_05
		jmp	libmio_tvc		;standard mode
libmio_iss_05:	cpi	ZL,0x01			
		brne	libmio_iss_06		
		jmp	libmio_tvg2		;2 color graphics
libmio_iss_06:	cpi	ZL,0x02			
		brne	libmio_iss_07		
		jmp	libmio_tvg4		;4 color graphics
libmio_iss_07:	jmp	libmio_tvg8		;8 color graphics
				
		
;-------------------------------------------------------------------------------
; calculate sound
;-------------------------------------------------------------------------------
libmio_is_15:	ldi	YL,LOW(libmio_ram+10)
		ldi	YH,HIGH(libmio_ram+10)
		lds	tempreg1,libmio_rand2	;2 noise signal
		lds	XL,libmio_note_val	;2 note val
		cpi	XL,0xff
		breq	libmio_wa_01
		lds	XL,libmio_wpos_l	;2 wave pointer low
		lds	XH,libmio_wpos_h	;2 wave pointer high
		lds	r0,libmio_woff_l	;2 wave offset low
		lds	r1,libmio_woff_h	;2 wave offset high
		add	XL,r0			;1 calculate new pointer
		adc	XH,r1			;1 
		sts	libmio_wpos_l,XL	;2 writeback new pointer
		sts	libmio_wpos_h,XH	;2 
		ldi	ZH,HIGH(libmio_wtable*2)	;1 start wavetable 1 high
		ldi	ZL,LOW(libmio_wtable*2)	;1 start wavetable 1 low
		add	ZL,XH			;1 wave-position
		lpm	tempreg1,Z		;3 get wave value
libmio_wa_01:	lds	tempreg2,libmio_envv	;2 envelope value
		mulsu	tempreg1,tempreg2	;2 shaped value
		ldi	XH,0x80			;1 neutral signal
		add	XH,r1			;1 add result
		sts	libmio_pwm,XH		;2 set level
		
;-------------------------------------------------------------------------------
; pseudo-random generator
;-------------------------------------------------------------------------------
		lds	ZL,libmio_rand1
		lds	ZH,libmio_rand2
		lds	tempreg1,libmio_rand3
		lds	tempreg2,libmio_rand4
		clr	XL
		sbrc	ZL,2
		com	XL
		sbrc	ZL,5
		com	XL
		sbrc	ZH,4
		com	XL
		sbrc	tempreg2,7
		com	XL
		rol	XL
		rol	ZL
		rol	ZH
		rol	tempreg1
		rol	tempreg2
		sts	libmio_rand1,ZL
		sts	libmio_rand2,ZH
		sts	libmio_rand3,tempreg1
		sts	libmio_rand4,tempreg2

;------------------------------------------------------------------------------
;the keyboard routine
;------------------------------------------------------------------------------
libmio_kir:	lds	ZL,UCSR0A		;status
		sbrs	ZL,RXC0			;RXC?
		rjmp	libmio_kbd_12		;no byte available
		lds	ZL,UDR0		

		lds	XL,libmio_kbdstate	;get key state
		cpi	ZL,0xf0			;key release
		brne	libmio_kbd_01		;jump if not
		ori	XL,0x80			;set release flag
		rjmp	libmio_kbd_11		;go to the end

libmio_kbd_01:	cpi	ZL,0xe0			;extended code
		brne	libmio_kbd_01a		;jump if not
		ori	XL,0x40			;set extended flag
		rjmp	libmio_kbd_11		;thats all

libmio_kbd_01a:	cpi	ZL,0x83			;extended code
		brne	libmio_kbd_02		;jump if not
		ldi	ZL,0x10

libmio_kbd_02:	sbrc	ZL,7			;skip if no bit 7
		rjmp	libmio_kbd_12		;thats all

		cpi	ZL,0x12			;left shift
		brne	libmio_kbd_03		;no
		ori	XL,0x01			;set shift bit
		sbrc	XL,7			;skip if not release
		andi	XL,0xfe			;clear shift bit
		rjmp	libmio_kbd_10		;thats all
		
libmio_kbd_03:	cpi	ZL,0x59			;right shift
		brne	libmio_kbd_04		;jump if not
		ori	XL,0x02			;set shift bit
		sbrc	XL,7			;skip if not release
		andi	XL,0xfd			;clear shift bit
		rjmp	libmio_kbd_10		;thats all

libmio_kbd_04:	cpi	ZL,0x14			;control
		brne	libmio_kbd_05		;jump if not
		sbrc	XL,6			;extended? (right ctrl key)
		rjmp	libmio_kbd_04a
		ori	XL,0x04			;set control bit
		sbrc	XL,7			;skip if not release
		andi	XL,0xfb			;clear ctrl bit
		rjmp	libmio_kbd_10		;thats all
libmio_kbd_04a:	ori	XL,0x08			;set control bit
		sbrc	XL,7			;skip if not release
		andi	XL,0xf7			;clear ctrl bit
		rjmp	libmio_kbd_10		;thats all
		
libmio_kbd_05:	cpi	ZL,0x11			;alternate
		brne	libmio_kbd_06		;jump if not
		ori	XL,0x10			;set alt bit
		sbrc	XL,7			;skip if not release
		andi	XL,0xef			;clear alt bit
		rjmp	libmio_kbd_10		;thats all
		
libmio_kbd_06:	sbrs	XL,7			;check release flag
		rjmp	libmio_kbd_06a		
		sts	libmio_keycode,byte0	;clear keycode
		rjmp	libmio_kbd_10		

libmio_kbd_06a:	sts	libmio_scancode,ZL	;var for scancode
		ldi	ZH,HIGH(libmio_ktables)	;start ext keytable 
		lds	XH,libmio_ksel		;select keytable
		add	ZH,XH			;offset
		lsl	ZH			;*2 for lpm
;check for extended		
		sbrs	XL,6			;skip if extended
		rjmp	libmio_kbd_06b		;->next check
		inc	ZH			;table 3 (extended)
		andi	ZL,0x7f			;mask bit 7
		rjmp	libmio_kbd_08
;check for alt
libmio_kbd_06b:	sbrs	XL,4			;skip if alt
		rjmp	libmio_kbd_06c		;->next check
		inc	ZH			;table 4 (alt)
		ori	ZL,0x80			;mask bit 7
		rjmp	libmio_kbd_08
		

libmio_kbd_06c:	andi	ZL,0x7f			;no shift
		sbrc	XL,0			;left shift active
		ori	ZL,0x80			;part 2 of table
		sbrc	XL,1			;right shift active
		ori	ZL,0x80			;part 2 of table
libmio_kbd_08:	lpm	ZL,Z			;get value
		lds	ZH,libmio_scancode	;scancode
libmio_kbd_8a:	sts	libmio_keycode,ZL	;keycode
		sts	libmio_lastkey,ZL	;lastkey
		cpi	ZL,0xe9			;del?
		brne	libmio_kbd_09
		sbrs	XL,2			;control?
		rjmp	libmio_kbd_09
		sbrs	XL,4			;alternate?
		rjmp	libmio_kbd_09
		jmp	start			;restart

libmio_kbd_09:	ldd	tempreg1,Y+27		;keyboard control
		sbrs	XL,2			;left control?
		rjmp	libmio_kbd_09b
		cpi	ZL,'c'			;Break?
		brne	libmio_kbd_09a
		ori	tempreg1,1		;set break bit
libmio_kbd_09a:	cpi	ZL,'p'			;Screenshot?
		brne	libmio_kbd_09b
		ori	tempreg1,2		;set screenshot bit
libmio_kbd_09b:	sbrs	XL,3			;right control?
		rjmp	libmio_kbd_09h
		cpi	ZL,'c'			;Break?
		brne	libmio_kbd_09c
		ori	tempreg1,1		;set break bit
libmio_kbd_09c:	cpi	ZL,'p'			;Screenshot?
		brne	libmio_kbd_09h
		ori	tempreg1,2		;set screenshot bit

libmio_kbd_09h:	cpi	ZL,0xfc			;monitor?
		brne	libmio_kbd_09i
		ori	tempreg1,4		;set monitor bit				
libmio_kbd_09i:	std	Y+27,tempreg1		;save keyboard control
		
libmio_kbd_10: 	andi	XL,0x3f			;clear ext and release bit
libmio_kbd_11:	sts	libmio_kbdstate,XL	;store keyboard status
libmio_kbd_12:

;-------------------------------------------------------------------------------
;serial port (receive)
;-------------------------------------------------------------------------------		
libmio_is_17:	lds	ZH,libmio_serio		;get signal byte
		lds	XL,libmio_rcnt		;clock counter
		cpi	XL,0x00			;end
		breq	libmio_is_18
		cpi	XL,0xff			;wait for sig
		brne	libmio_ser_01
		sbrc	ZH,0			;skip if line is 0
		rjmp	libmio_is_18
		ldi	XL,124			;start value
		sts	libmio_rcnt,XL
		rjmp	libmio_is_18
libmio_ser_01:	cpi	XL,105
		breq	libmio_ser_02		;bit 0		
		cpi	XL,92
		breq	libmio_ser_02		;bit 1		
		cpi	XL,79
		breq	libmio_ser_02		;bit 2
		cpi	XL,66
		breq	libmio_ser_02		;bit 3
		cpi	XL,53
		breq	libmio_ser_02		;bit 4
		cpi	XL,40
		breq	libmio_ser_02		;bit 5
		cpi	XL,27
		breq	libmio_ser_02		;bit 6
		cpi	XL,14
		breq	libmio_ser_02		;bit 7
		dec	XL
		sts	libmio_rcnt,XL
		rjmp	libmio_is_18
		
libmio_ser_02:	dec	XL
		sts	libmio_rcnt,XL
		lds	XL,libmio_rdat		;serial data
		lsr	XL
		bst	ZH,0			;RXD data
		bld	XL,7			;shift in
		sts	libmio_rdat,XL
		
;-------------------------------------------------------------------------------
;serial port (send)
;-------------------------------------------------------------------------------		
libmio_is_18:	lds	XL,libmio_tcnt		;clock counter
		cpi	XL,0x00	
		breq	libmio_is_20		;nothing to do
		cpi	XL,0xff			;wait for start
		brne	libmio_set_01
		ldi	XL,144
		sts	libmio_tcnt,XL
		andi	ZH,0x7f			;clear bit		
		rjmp	libmio_set_04
		
libmio_set_01:	cpi	XL,131
		breq	libmio_set_03		;bit 0
		cpi	XL,118
		breq	libmio_set_03		;bit 1
		cpi	XL,105
		breq	libmio_set_03		;bit 2
		cpi	XL,92
		breq	libmio_set_03		;bit 3
		cpi	XL,79
		breq	libmio_set_03		;bit 4
		cpi	XL,66
		breq	libmio_set_03		;bit 5
		cpi	XL,53
		breq	libmio_set_03		;bit 6
		cpi	XL,40
		breq	libmio_set_03		;bit 7
		cpi	XL,27
		brne	libmio_set_02		;stop bit
		ldi	ZH,0xff
libmio_set_02:	dec	XL
		sts	libmio_tcnt,XL
		rjmp	libmio_set_04

libmio_set_03:	lds	XH,libmio_tdat
		bst	XH,0
		lsr	XH
		bld	ZH,7
		sts	libmio_tdat,XH
		dec	XL
		sts	libmio_tcnt,XL
libmio_set_04:	sts	libmio_serio,ZH

;-------------------------------------------------------------------------------
;restore registers and exit
;-------------------------------------------------------------------------------		
libmio_is_20:	pop	ZH			;2 restore sreg
		out	SREG,ZH			;1 put back    
		pop	r1			;2 restore r1
		pop	r0			;2 restore r0
		pop	XL			;2 restore Y-register
		pop	XH			;2 restore Y-register
libmio_is_30:	pop	YL			;2 restore Y-register
		pop	YH			;2 restore Y-register
		pop	ZL			;2 restore Z-register
		pop	ZH			;2 restore Z-register
		pop	tempreg3		;
		pop	tempreg2		;2
		pop	tempreg1		;2 restore register
		reti				;4 thats all

;-------------------------------------------------------------------------------
;port konfiguration for port B ... D
;-------------------------------------------------------------------------------
libmio_init:	ldi	XL,0xba			;port direction (+SPI)
		out	DDRB,XL
		ldi	XL,0xff
		out	PORTB,XL	

		ldi	XL,0xe0			;port direction
		out	DDRC,XL
		ldi	XL,0x00
		out	PORTC,XL	

		ldi	XL,0xf0			;set pin 4-7 as output
		out	DDRD,XL
		ldi	XL,0xff
		out	PORTD,XL	

		in	XL,PINC			;get portc (bit 2-4 = config)
		andi	XL,0x7f			;clear editor mode bit
		sts	libmio_config,XL	;store configuration

		
;-------------------------------------------------------------------------------
;timer configuration
;-------------------------------------------------------------------------------
		ldi	XH,0xc2			;FAST-PWM-Mode 14 (PWM channel B)
		sts	TCCR1A,XH
		ldi	XH,0x19			;FAST-PWM-Mode (no prescaler)
		sts	TCCR1B,XH
		ldi	ZH,HIGH(libmio_counter_p-1)
		sbrs	XL,2			;1 pal/ntsc bit
		ldi	ZH,HIGH(libmio_counter_n-1)
		sts	ICR1H,ZH
		ldi	ZH,LOW(libmio_counter_p-1)
		sbrs	XL,2			;1 pal/ntsc bit
		ldi	ZH,LOW(libmio_counter_n-1)
		sts	ICR1L,ZH

		sts	OCR1AH,byte0		;we use only 8 bits
		ldi	ZH,libmio_hslen	;1 hsync length pal
		sts	OCR1AL,ZH		

		sts	OCR1BH,byte0		;we use only 8 bits
		ldi	ZH,libmio_hstart	;1 hor position pal
		sts	OCR1BL,ZH		
				
		ldi	XL,0x04
		sts	TIMSK1,XL		;enable compare OC1B int

		ldi	XL,0x02			;ctc but disable output
		out	TCCR0A,XL
		ldi	XL,0x01
		out	TCCR0B,XL
		ldi	XL,49			;ca. 200KHz
		out	OCR0A,XL			

		ldi	XL,0x83			;timer 2 config
		sts	TCCR2A,XL
		ldi	XL,0x01			;timer 2 config
		sts	TCCR2B,XL
		ldi	XL,0x80			;zero signal
		sts	OCR2A,XL
		sts	ASSR,byte0

;-------------------------------------------------------------------------------
;usart configuration (synchronous for keyboard)
;-------------------------------------------------------------------------------
		sts	UCSR0A,byte0
		ldi	XL,0x10
		sts	UCSR0B,XL
		ldi	XL,0x66
		sts	UCSR0C,XL

;-------------------------------------------------------------------------------
;spi configuration (mode 3 master 10MHz)
;-------------------------------------------------------------------------------
		ldi	XL,0x5c
		out	SPCR0,XL
		ldi	XL,0x01
		out	SPSR0,XL

;-------------------------------------------------------------------------------
;clear video-RAM and set pointers
;-------------------------------------------------------------------------------
		ldi	XL,0x15			;color = white
		rcall	libmio_setcol		;set color
		ldi	YL,LOW(libmio_ram)
		ldi	YH,HIGH(libmio_ram)
		ldi	XL,LOW(libmio_vram)	;2
		st	Y+,XL			;2 start vram
		ldi	XL,HIGH(libmio_vram)	;2
		st	Y+,XL			;2 start vram
		ldi	XL,0x1e
libmio_init_c:	st	Y+,byte0
	    	dec	XL
		brne	libmio_init_c	
		ldi	XL,0xaa	    
		sts	libmio_rand1,XL
		sts	libmio_rand3,XL
		sts	libmio_channel,byte0	;set output to screen

;------------------------------------------------------------------------------
; set display to normal mode
;------------------------------------------------------------------------------
libmio_nmode:	push	r16
		lds	r16,libmio_config
		andi	r16,0x7f	
		sts	libmio_config,r16
		pop	r16
		ret		

;------------------------------------------------------------------------------
; set display to editor mode
;------------------------------------------------------------------------------
libmio_emode:	push	r16
		lds	r16,libmio_config
		ori	r16,0x80    	
		sts	libmio_config,r16
		pop	r16
		ret		
		
;------------------------------------------------------------------------------
; set color to XL
;------------------------------------------------------------------------------
libmio_setcol:	lsl	XL
		andi	XL,0xee
		sts	libmio_color,XL
		ret		

;------------------------------------------------------------------------------
; write char tempreg1 to Y
;------------------------------------------------------------------------------
libmio_wchar:	st	Y,tempreg1
		lds	tempreg1,libmio_config
		sbrc	tempreg1,7
		rjmp	libmio_wchar1
		push	YL
		push	YH
		push	tempreg2
		ldi	tempreg1,LOW(libmio_cols*libmio_rows)
		add	YL,tempreg1
		ldi	tempreg1,HIGH(libmio_cols*libmio_rows)
		adc	YH,tempreg1		
		lds	tempreg1,libmio_color
		andi	tempreg1,0xee
		st	Y,tempreg1
		pop	tempreg2
		pop	YH
		pop	YL
libmio_wchar1:	ret

;-------------------------------------------------------------------------------
; Bin16 -> ASCII (dez)
; XL,XH 	= number
; ctrl		= format
; 	Bit 0,1	 = format	00 = compact
;				01 = leading spaces
;				10 = leading zeroes 0...65536
;				11 = leading zeroes with +- sign  
;			
;	Bit 2,3	 = number of visible digits (2,3,4,5)
;	Bit 4,5	 = comma position (no  xxxx,x   xxx,xx  xx,xxx) 
;	Bit 6	1= big size (only in monochrome mode) 
;-------------------------------------------------------------------------------
.ifdef use_libmio_outd
libmio_outd:	push	tempreg1		;save registers
		push	tempreg2
		push	tempreg3
		push	XL
		push	XH
		ldi	tempreg1,0x20		;space
		mov	tempreg2,ctrl
		andi	tempreg2,0x03
		cpi	tempreg2,0x02		;leading zeroes without sign
		breq	libmio_outd_03
		cpi	tempreg2,0x03		;leading zeroes with +- sign
		brne	libmio_outd_01
		ldi	tempreg1,'+'		;set positive sign
libmio_outd_01:	sbrs	XH,7
		rjmp	libmio_outd_02
		com	XL			;X=-X
		com	XH			;
		add	XL,byte1		;
		adc	XH,byte0		;
		ldi	tempreg1,'-'		;sign is negative
libmio_outd_02:	call	libmio_char		;out sign


libmio_outd_03:	mov	tempreg2,ctrl		;format byte
		andi	tempreg2,0x0c		;mask number of digits
		mov	tempreg3,ctrl		;format byte
		andi	tempreg3,0x30		;mask comma position
		clt				;clear nonzero-flag
		ldi	tempreg1,47		;"0"-1 	
libmio_outd_04:	inc	tempreg1			;
		subi	XL,16			;-10000
		sbci	XH,39
		brcc	libmio_outd_04		;no overflow
		subi	XL,240			;+10000
		sbci	XH,216			;
		cpi	tempreg2,0x0c		;5 digits?
		brne	libmio_outd_05		;no		
		rcall	libmio_nc		;out digit 10^4

libmio_outd_05:	ldi	tempreg1,47		;"0"-1 	
libmio_outd_06:	inc	tempreg1			;
		subi	XL,232			;-1000
		sbci	XH,3			;
		brcc	libmio_outd_06		;no overflow
		subi	XL,24			;+1000
		sbci	XH,252			;
		sbrc	tempreg2,3		;skip if < 4 digits
		rcall	libmio_nc		;out digit 10^3
		cpi	tempreg3,0x30		;comma?
		brne	libmio_outd_07		;no
		rcall	libmio_comma		;

libmio_outd_07:	ldi	tempreg1,47		;"0"-1 	
libmio_outd_08:	inc	tempreg1		;+1
		subi	XL,100			;-100
		sbc	XH,byte0		;
		brcc	libmio_outd_08		;no overflow
		subi	XL,156			;+100
		sbci	XH,255			;
		cpi	tempreg2,0x00		;3 digits
		breq	libmio_outd_09
		rcall	libmio_nc		;out digit 10^2
libmio_outd_09:	cpi	tempreg3,0x20		;comma?
		brne	libmio_outd_10		;no
		rcall	libmio_comma		;

libmio_outd_10:	ldi	tempreg1,47		;"0"-1 	
libmio_outd_11:	inc	tempreg1		;+1
		subi	XL,10			;-10
		brcc	libmio_outd_11		;no overflow
		rcall	libmio_nc		;out digit 10^1
		subi	XL,246			;+10
		cpi	tempreg3,0x10		;comma?
		brne	libmio_outd_12		;no
		rcall	libmio_comma		;
		
libmio_outd_12:	ldi	tempreg1,48		;"0"
		add	tempreg1,XL		;LSB digit
		rcall	libmio_char
		pop	XH
		pop	XL
		pop	tempreg3
		pop	tempreg2
		pop	tempreg1
		ret

;-------------------------------------------------------------------------------
; out numeric char (tempreg1)
;-------------------------------------------------------------------------------
libmio_nc:	cpi	tempreg1,48		;is zero
		breq	libmio_nc_1		;yes
		set				;set flag
		rjmp	libmio_char		;send char
		
libmio_nc_1:	sbrc	ctrl,1			;skip if no leading zeroes
libmio_nc_2:	rjmp	libmio_char		;send zero
		brts	libmio_nc_2		;flag is set
		
		sbrs	ctrl,0			;skip if no leading spaces
		ret 
		rjmp	libmio_space		;send space
		ret
.endif

;-------------------------------------------------------------------------------
; Bin -> ASCII (hex) 
; (XL,XH) = number
; ctrl 		= format
; 	Bit 0	1= 16Bit (4 chars)
;	Bit 6	1= Big size
;-------------------------------------------------------------------------------
.ifdef use_libmio_outh
libmio_outh:	push	tempreg1		;save temp register
		sbrs	ctrl,0
		rjmp	libmio_outh1		;only 8
		mov	tempreg1,XH		;
		swap	tempreg1		;high-nibble
		rcall	libmio_oh
		mov	tempreg1,XH		;
		rcall	libmio_oh
libmio_outh1:	mov	tempreg1,XL		;
		swap	tempreg1		;high-nibble
		rcall	libmio_oh
		mov	tempreg1,XL		;
		rcall	libmio_oh
		pop	tempreg1		;restore temp register
		ret

;-------------------------------------------------------------------------------
; out hexadecimal char (tempreg1)
;-------------------------------------------------------------------------------
libmio_oh:	andi	tempreg1,0x0f		;mask nibble
		subi	tempreg1,0xd0		;+30
		cpi	tempreg1,0x3a		;>9
		brcs	libmio_oh_1		;jump, if not
		subi	tempreg1,0xf9		;+6
libmio_oh_1:	rjmp	libmio_char		
.endif
		
;-------------------------------------------------------------------------------
; clear video-RAM
;-------------------------------------------------------------------------------
libmio_clr:	lds	tempreg1,libmio_vidmode	;video mode
		cpi	tempreg1,0x00		;standard?
		breq	libmio_fill		;yes
		push	XL
		push	XH
		push	YL
		push	YH
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		ldi	XL,libmio_rows*4
libmio_cls_g01:	ldi	XH,libmio_cols
libmio_cls_g02:	st	Y+,byte0		;write to RAM
		dec	XH			;decrement counter
		brne	libmio_cls_g02		;loop
		dec	XL			;decrement counter
		brne	libmio_cls_g01		;loop
		pop	YH
		pop	YL
		clr	libmio_cur_x		;set X-position
		clr	libmio_cur_y		;set Y-position
		sts	libmio_lpos,byte0
		pop	XH
		pop	XL
		ret
		
		ldi	tempreg1,32		;space
libmio_fill:	push	XL
		push	XH
		push	YL
		push	YH
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		ldi	XL,libmio_rows
libmio_cls_01:	ldi	XH,libmio_cols
libmio_cls_02:	st	Y+,tempreg1		;write to RAM
		dec	XH			;decrement counter
		brne	libmio_cls_02		;loop
		dec	XL			;decrement counter
		brne	libmio_cls_01		;loop
		lds	tempreg1,libmio_color
		andi	tempreg1,0xee
		ldi	XL,libmio_rows
libmio_cls_05:	ldi	XH,libmio_cols
libmio_cls_06:	st	Y+,tempreg1		;write to RAM
		dec	XH			;decrement counter
		brne	libmio_cls_06		;loop
		dec	XL			;decrement counter
		brne	libmio_cls_05		;loop
		pop	YH
		pop	YL
libmio_cls_03:	clr	libmio_cur_x		;set X-position
		clr	libmio_cur_y		;set Y-position
libmio_cls_04:	sts	libmio_lpos,byte0
		pop	XH
		pop	XL
		ret				;the end

;-------------------------------------------------------------------------------
; newline and calculate new position
;-------------------------------------------------------------------------------
.ifdef use_libmio_newline
libmio_newline:	push	XL
		push	XH
		sts	libmio_lpos,byte0
		lds	XL,libmio_channel
		cpi	XL,0			;screen
		brne	libmio_newl1
		ldi	XL,libmio_rows
		clr	libmio_cur_x		;begin of line
		inc	libmio_cur_y		;next line
		cp	XL,libmio_cur_y		;beyond screen
		brne	libmio_cls_04
		ldi	XH,(libmio_rows-1)	;last line
		mov	libmio_cur_y,XH
		push	YL
		push	YH
		rjmp	libmio_char_04		;scroll screen

libmio_newl1:	cpi	XL,1			;serial
		brne	libmio_newl2
		pop	XH
		pop	XL
		ldi	tempreg1,0x0a		;newline
		jmp	libmio_pser		;write serial
		
libmio_newl2:	cpi	XL,2			;parallel
		brne	libmio_newl3
		pop	XH
		pop	XL
		ldi	tempreg1,0x0a		;newline
		rjmp	libmio_ppar		;write parallel
		
libmio_newl3:	cpi	XL,3			;array
		brne	libmio_newl4
		pop	XH
		pop	XL
		ret

libmio_newl4:	pop	XH			;i2c
		pop	XL
		ldi	tempreg2,0x0a
		libi2c_wbyte
		ret
.endif

;-------------------------------------------------------------------------------
; output character and calculate new position (tempreg1=char)
;-------------------------------------------------------------------------------
libmio_comma:	ldi	tempreg1,0x2c		;set char to comma
		rjmp	libmio_char		;write
libmio_space:	ldi	tempreg1,32		;set char to space
libmio_char:	push	XL
		push	XH
		lds	XL,libmio_lpos
		inc	XL
		sts	libmio_lpos,XL
		lds	XL,libmio_channel
		cpi	XL,4
		brcs	libmio_chr_1
		pop	XH
		pop	XL
		mov	tempreg2,tempreg1	;copy char
		libi2c_wbyte			;output to i2c
		ret
		
libmio_chr_1:	cpi	XL,0			;screen
		breq	libmio_char2
		cpi	XL,3			;array
		brne	libmio_char0
		mov	XL,tempreg1
		clr	XH
		push	tempreg2
		push	tempreg3
		call	bas_warr		;write to array
		pop	tempreg3
		pop	tempreg2
		pop	XH
		pop	XL
		ret	
		
libmio_char0:	pop	XH
		pop	XL
		lds	XL,libmio_channel
		sbrs	XL,1			;serial
		jmp	libmio_pser		;write serial
libmio_char1:	rjmp	libmio_ppar		;write parallel

libmio_char2:	push	ZL			;save
		lds	ZL,libmio_vidmode	;get video mode
		cpi	ZL,0x00
		breq	libmio_char_0
		rjmp	libmio_gchar		;graphics mode
		
libmio_char_0:	pop	ZL
		push	YL
		push	YH
		rcall	libmio_calc		;calculate address		
		sbrc	ctrl,6			;big chars
		rjmp	libmio_bchar

		rcall	libmio_wchar		;write char
		inc	XL			;next position
		cpi	XL,libmio_cols		;beyond the end of line
		brne	libmio_char_01
		clr	XL			;begin of line
		inc	XH			;of next line
		cpi	XH,libmio_rows		;beyond screen
		brne	libmio_char_01
		ldi	XH,libmio_rows-1	;last line
		rjmp	libmio_char_03		;we must scroll			
libmio_char_01:	mov	libmio_cur_x,XL		;X
		mov	libmio_cur_y,XH		;Y
libmio_char_02:	pop	YH
		pop	YL
		pop	XH
		pop	XL
		ret

; scroll screen
libmio_char_03:	mov	libmio_cur_x,XL		;save X-position
		mov	libmio_cur_y,XH		;save Y-position
libmio_char_04:	ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		push	tempreg1
		ldi	XL,libmio_rows-1	;lines -1
libmio_char_05:	ldi	XH,libmio_cols		
libmio_char_06:	ldd	tempreg1,Y+libmio_cols
		st	Y+,tempreg1
		dec	XH
		brne	libmio_char_06
		dec	XL
		brne	libmio_char_05		
		ldi	XL,libmio_cols
		ldi	tempreg1,32		;space
libmio_char_07:	st	Y+,tempreg1
		dec	XL
		brne	libmio_char_07

libmio_char_14:	ldi	YH,HIGH(libmio_vram+libmio_cols*libmio_rows)	;attr start
		ldi	YL,LOW(libmio_vram+libmio_cols*libmio_rows)	;attr start

		ldi	XL,libmio_rows-1	;lines -1
libmio_char_15:	ldi	XH,libmio_cols		
libmio_char_16:	ldd	tempreg1,Y+libmio_cols
		st	Y+,tempreg1
		dec	XH
		brne	libmio_char_16
		dec	XL
		brne	libmio_char_15		
		ldi	XL,libmio_cols
		lds	tempreg1,libmio_color	;space
libmio_char_17:	st	Y+,tempreg1
		dec	XL
		brne	libmio_char_17
		pop	tempreg1
		rjmp	libmio_char_02


libmio_bchar:	cpi	XL,libmio_cols-3
		brcc	libmio_char_02
		cpi	XH,libmio_rows-5
		brcc	libmio_char_02
		subi	XL,0xfd			;X+3
		mov	libmio_cur_x,XL		;store new X
		push	ZL
		push	ZH
		push	tempreg5		;byte 1
		push	tempreg6		;byte 2
		push 	tempreg7		;byte 3
		push	tempreg8		;for stroring T-bit
		bld	tempreg8,0		;store T
		push	tempreg2		;temp value

libmio_bchar_0:	mov	ZL,tempreg1		;char
		ldi	ZH,HIGH(libmio_ctable*2)	;start char ROM
		ldi	XH,5			;10 char lines
		
libmio_bchar_1:	clr	tempreg5		;clear columns
		clr	tempreg6
		clr	tempreg7
		clr	tempreg2		;clear temp value
		
		lpm	XL,Z			;get pixel line
		sbrc	XL,7			;pixel 0
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg5,0
		sbrc	XL,6			
		inc	tempreg2
		push 	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg5,1
		sbrc	XL,5			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg6,0
		sbrc	XL,3			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg6,1
		sbrc	XL,2			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg7,0
		pop	tempreg2
		sbrc	XL,1			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg7,1

		inc	ZH			;next line
		clr	tempreg2		;clear temp value
		
		lpm	XL,Z			;get pixel line
		sbrc	XL,7			;pixel 0
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg5,2
		sbrc	XL,6			
		inc	tempreg2
		push 	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg5,3
		sbrc	XL,5			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg6,2
		sbrc	XL,3			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg6,3
		sbrc	XL,2			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg7,2
		pop	tempreg2
		sbrc	XL,1			
		inc	tempreg2
		bst	tempreg2,0		;copy pix
		bld	tempreg7,3

libmio_bchar_d:	inc	ZH
		mov	tempreg1,tempreg5	;char1
		rcall	libmio_wchar
		adiw	YL,0x01
		mov	tempreg1,tempreg6	;char2
		rcall	libmio_wchar
		adiw	YL,0x01
		mov	tempreg1,tempreg7	;char3
		rcall	libmio_wchar
		adiw	YL,libmio_cols-2	;next line
		dec	XH
		breq	libmio_bchar_e
		rjmp	libmio_bchar_1
		
libmio_bchar_e:	pop	tempreg2
		bst	tempreg8,0		;restore T
		pop	tempreg8
		pop	tempreg7
		pop	tempreg6
		pop	tempreg5
		pop	ZH
		pop	ZL
		rjmp	libmio_char_02

;graphics mode
libmio_gchar:	push	ZH			;restore Z reg
		push	YL
		push	YH
		push	tempreg3
		mov	ZL,tempreg1		;char
		ldi	ZH,HIGH(libmio_ctable*2)	;start char ROM
		mov	XL,libmio_cur_x		;get X-position
		mov	XH,libmio_cur_y		;get Y-position
		ldi	YH,10			;number of lines
libmio_gchar1:	lpm	YL,Z			;get pixline
		rcall	libmio_plpl
		inc	XH
		mov	XL,libmio_cur_x		;get X-position
		sbrc	ctrl,6			;bigchars
		rcall	libmio_plpl
		sbrc	ctrl,6			;bigchars
		inc	XH
		inc	ZH			;next pixline
		mov	XL,libmio_cur_x		;get X-position
		dec	YH
		brne	libmio_gchar1		
		mov	XL,libmio_cur_x		;get X-position
		subi	XL,0xfa			;+6
		sbrc	ctrl,6
		subi	XL,0xfa			;+6
		mov	libmio_cur_x,XL		
		pop	tempreg3
		pop	YH
		pop	YL
		pop	ZH		
		pop	ZL
		pop	XH
		pop	XL
		ret

;-------------------------------------------------------------------------------
; plot pixline in graphics mode, YL=pixline
;-------------------------------------------------------------------------------
libmio_plpl:	lds	tempreg3,libmio_color
		push	tempreg3
		swap	tempreg3
		sbrc	YL,7			;pix 1
		swap	tempreg3		
		sts	libmio_color,tempreg3
		rcall	libmio_plot		;output pix
		inc	XL			;next pos
		sbrc	ctrl,6			;no bigchar
		rcall	libmio_plot		;output pix again
		sbrc	ctrl,6			;no bigchar
		inc	XL			;next pos
		
		sbrc	YL,6			;pix 2
		swap	tempreg3		
		push	tempreg3
		sts	libmio_color,tempreg3
		rcall	libmio_plot		;output pix
		inc	XL			;next pos
		sbrc	ctrl,6			;no bigchar
		rcall	libmio_plot		;output pix again
		sbrc	ctrl,6			;no bigchar
		inc	XL			;next pos

		sbrc	YL,5			;pix 3
		swap	tempreg3		
		sts	libmio_color,tempreg3
		rcall	libmio_plot		;output pix
		inc	XL			;next pos
		sbrc	ctrl,6			;no bigchar
		rcall	libmio_plot		;output pix again
		sbrc	ctrl,6			;no bigchar
		inc	XL			;next pos

		sbrc	YL,3			;pix 4
		swap	tempreg3		
		sts	libmio_color,tempreg3
		rcall	libmio_plot		;output pix
		inc	XL			;next pos
		sbrc	ctrl,6			;no bigchar
		rcall	libmio_plot		;output pix again
		sbrc	ctrl,6			;no bigchar
		inc	XL			;next pos

		sbrc	YL,2			;pix 5
		swap	tempreg3		
		sts	libmio_color,tempreg3
		rcall	libmio_plot		;output pix
		inc	XL			;next pos
		sbrc	ctrl,6			;no bigchar
		rcall	libmio_plot		;output pix again
		sbrc	ctrl,6			;no bigchar
		inc	XL			;next pos

		pop	tempreg3
		sbrc	YL,1			;pix 6
		swap	tempreg3		
		sts	libmio_color,tempreg3
		rcall	libmio_plot		;output pix
		inc	XL			;next pos
		sbrc	ctrl,6			;no bigchar
		rcall	libmio_plot		;output pix again
		sbrc	ctrl,6			;no bigchar
		inc	XL			;next pos

		pop	tempreg3		;restore color
		sts	libmio_color,tempreg3
		ret
			
;-------------------------------------------------------------------------------
; output text from ROM (Z=text,nullterminated)
;-------------------------------------------------------------------------------
.ifdef use_libmio_orom
libmio_orom:	push	tempreg1	
libmio_orom_1:	lpm	tempreg1,Z+		;get char
		cpi	tempreg1,0		;end?
		brne	libmio_orom_2		;no
		pop	tempreg1
		ret				;the end
libmio_orom_2:	rcall	libmio_char		;write char to screen
		rjmp	libmio_orom_1		;next char	
.endif

;-------------------------------------------------------------------------------
; output text from ROM (following y,x,nullterminated text)
;-------------------------------------------------------------------------------
.ifdef use_libmio_othis
libmio_othis:	pop	ZH			;get return address
		pop	ZL
		lsl	ZL			;*2 so we use bytes
		rol	ZH
		push	XL
		push	XH
		lpm	XH,Z+			;get first value (255/Y)
		cpi	XH,0xff			;no position
		breq	libmio_othis_0
		lpm	XL,Z+			;get second value
		rcall	libmio_gxy		;set position
libmio_othis_0:	pop	XH
		pop	XL
		push	tempreg1		;save reg
libmio_othis_1:	lpm	tempreg1,Z+		;get char
		cpi	tempreg1,0		;end?
		brne	libmio_othis_2		;no
		pop	tempreg1		;restore reg
		sbiw	ZL,1			;point to zero
		lsr	ZH			;/2
		ror	ZL
		adiw	ZL,1			;point to next instruction
		ijmp
libmio_othis_2:	rcall	libmio_char		;write char to screen
		rjmp	libmio_othis_1		;next char	
.endif

;-------------------------------------------------------------------------------
; output text from RAM (Z=text,nullterminated)
;-------------------------------------------------------------------------------
.ifdef use_libmio_oram
libmio_oram:	push	tempreg1
libmio_oram_1:	ld	tempreg1,Z+		;get char
		cpi	tempreg1,0		;end?
		brne	libmio_oram_2		;no
		pop	tempreg1
		ret				;the end
libmio_oram_2:	call	libmio_char		;write char to screen
		rjmp	libmio_oram_1		;next char	
.endif

;-------------------------------------------------------------------------------
; output text (multiple char) ZL=number
;-------------------------------------------------------------------------------
.ifdef use_libmio_mtext
libmio_mtext:	call	libmio_char		;write char to screen
		dec	ZL
		brne	libmio_mtext	
		ret
.endif

;-------------------------------------------------------------------------------
; set cursor to position XL:XH
;-------------------------------------------------------------------------------
.ifdef use_libmio_gotoxy
libmio_gxy:	push	ZL
		lds	ZL,libmio_vidmode
		cpi	ZL,0x00
		breq	libmio_gxy_00
		pop	ZL
		mov	libmio_cur_x,XL		;save X-position
		mov	libmio_cur_y,XH		;save Y-position
		ret

libmio_gxy_00:	pop	ZL		
		cpi	XL,libmio_cols
		brcs	libmio_gxy_01
		ldi	XL,libmio_cols-1
libmio_gxy_01:	cpi	XH,libmio_rows
		brcs	libmio_gxy_02
		ldi	XH,libmio_rows-1
libmio_gxy_02:	mov	libmio_cur_x,XL		;save X-position
		mov	libmio_cur_y,XH		;save Y-position
		ret
.endif

;-------------------------------------------------------------------------------
; plot  XL=X XH=Y  
;-------------------------------------------------------------------------------
.ifdef use_libmio_plot
libmio_plot:	push	YL
		push	YH
		lds	YL,libmio_clipx2	;right limit
		cp	YL,XL			;greater?
		brcs	libmio_nplot		;yes
		lds	YL,libmio_clipy2	;right limit
		cp	YL,XH			;greater?
		brcs	libmio_nplot		;yes
		rjmp	libmio_plot0		;OK, we can plot
libmio_nplot:	pop	YH			;restore registers
		pop	YL
		ret		
		
libmio_plot0:	push	XL			;save coords
		push	XH
		push	tempreg1
		push	tempreg2
		lds	YL,libmio_vidmode
		andi	YL,0x03
		breq	libmio_plot_0
		cpi	YL,0x02
		brcc	libmio_plot1
		rjmp	libmio_plot1_0
libmio_plot1:	brne	libmio_plot2
		rjmp	libmio_plot2_0
libmio_plot2:	rjmp	libmio_plot3_0				
	
libmio_plot_0:	ldi	tempreg2,1		;bit left top

		cpi	XL,libmio_cols*2	;limit X-coord
		brcc	libmio_plot_os
		cpi	XH,libmio_rows*2	;limit Y-coord
		brcc	libmio_plot_os

		sbrc	XH,0
		lsl	tempreg2
		sbrc	XH,0
		lsl	tempreg2
		sbrc	XL,0
		lsl	tempreg2		;tempreg2 = bitmask
		lsr	XH			;Y/2
		lsr	XL			;X/2
		ldi	YL,libmio_cols
		mul	XH,YL
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		add	YL,XL
		adc	YH,byte0
		add	YL,r0
		adc	YH,r1
		ldi	XL,LOW(libmio_cols*libmio_rows)
		ldi	XH,HIGH(libmio_cols*libmio_rows)
		add	YL,XL			;calc attr pointer
		adc	YH,XH
		ld	XL,Y			;attr at position
		lds	XH,libmio_color		;plot color
		andi	XH,0x0e			;only background is needed
		swap	XH			;swap f/b
		andi	XL,0xe0			;only background is needed
		cp	XH,XL			;background = plot color?
		breq	libmio_plot_r		;pixel must be only deleted

		lds	XH,libmio_color		;plot color
		andi	XH,0x0e			;only foreground
		or	XL,XH			;set new fg
		andi	XL,0xee			;font 0
		st	Y,XL			;write back attribute
		
libmio_plot_s:	ldi	XL,LOW(libmio_cols*libmio_rows)
		ldi	XH,HIGH(libmio_cols*libmio_rows)
		sub	YL,XL			;recalc char pointer
		sbc	YH,XH
		ld	XH,Y			;get char
		or	XH,tempreg2		;set pix
libmio_plot_e:	andi	XH,0x0f			;mask bits
		st	Y,XH				
libmio_plot_os:	pop	tempreg2		
		pop	tempreg1
		pop	XH
		pop	XL
		pop	YH
		pop	YL
		ret
	
libmio_plot_r:	ldi	XL,LOW(libmio_cols*libmio_rows)
		ldi	XH,HIGH(libmio_cols*libmio_rows)
		sub	YL,XL			;recalc char pointer
		sbc	YH,XH
		ld	XH,Y			;get char
		com	tempreg2
		and	XH,tempreg2		;clear pix
		rjmp	libmio_plot_e

libmio_plot1_0:	mov	YL,XL			;X
		andi	YL,0x07			;only bits
		inc	YL			;+1
		ldi	tempreg1,0x00		;bitmask
		sec				;set carry
libmio_plot1_1:	rol	tempreg1		;shift in bit
		dec	YL
		brne	libmio_plot1_1
;bit mask is in tempreg1
		lsr	XL			;X/8
		lsr	XL
		lsr	XL
		ldi	YL,libmio_v1cols/8
		mul	XH,YL
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		add	YL,XL
		adc	YH,byte0
		add	YL,r0
		adc	YH,r1
;address is now in Y	    	
		lds	XL,libmio_color		;active color
		lsr	XL
		andi	XL,0x01			;we check only bit 0
		brne	libmio_plot1_2		;set
		com	tempreg1
		ld	XL,Y
		and	XL,tempreg1
		st	Y,XL
		rjmp	libmio_plot_os
				
libmio_plot1_2:	ld	XL,Y
		or	XL,tempreg1
		st	Y,XL		
		rjmp	libmio_plot_os

libmio_plot2_0:	mov	YL,XL			;X
		lds	tempreg2,libmio_color	;active color
		lsr	tempreg2
		andi	tempreg2,0x03		;we check only bit 0+1
		ldi	tempreg1,0x03		;bitmask
		andi	YL,0x03			;only bits
libmio_plot2_1:	breq	libmio_plot2_2
		lsl	tempreg1
		lsl	tempreg1
		lsl	tempreg2
		lsl	tempreg2
		dec	YL
		brne	libmio_plot2_1
libmio_plot2_2:	com	tempreg1	
;bit mask is in tempreg1
;value is in tempreg2
		lsr	XL			;X/4
		lsr	XL
		ldi	YL,libmio_v2cols/4
		mul	XH,YL
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		add	YL,XL
		adc	YH,byte0
		add	YL,r0
		adc	YH,r1
;address is now in Y	    	
		ld	XL,Y
		and	XL,tempreg1
		or	XL,tempreg2
		st	Y,XL	
		rjmp	libmio_plot_os


libmio_plot3_0:	mov	YL,XL			;X
		lds	tempreg2,libmio_color	;active color
		lsr	tempreg2
		andi	tempreg2,0x07		;we check only bit 0-2
		ldi	tempreg1,0x07		;bitmask
		andi	YL,0x01			;only bits
		brne	libmio_plot3_1
		swap	tempreg1
		swap	tempreg2
libmio_plot3_1:	com	tempreg1	
;bit mask is in tempreg1
;value is in tempreg2
		lsr	XL			;X/2
		ldi	YL,libmio_v3cols/2
		mul	XH,YL
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		add	YL,XL
		adc	YH,byte0
		add	YL,r0
		adc	YH,r1
;address is now in Y	    	
		ld	XL,Y
		and	XL,tempreg1		;mask
		or	XL,tempreg2		;new bit
		st	Y,XL
		rjmp	libmio_plot_os
.endif


;-------------------------------------------------------------------------------
; correct char coordinates from YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
libmio_corrc:	cpi	YL,libmio_cols
		brcs	libmio_corrc1
		ldi	YL,libmio_cols-1
libmio_corrc1:	cpi	ZL,libmio_cols
		brcs	libmio_corrc2
		ldi	ZL,libmio_cols-1
libmio_corrc2:	cpi	YH,libmio_rows
		brcs	libmio_corrc3
		ldi	YH,libmio_rows-1
libmio_corrc3:	cpi	ZH,libmio_rows
		brcs	libmio_corrc4
		ldi	ZH,libmio_rows-1
libmio_corrc4:	cp	ZL,YL
		brcc	libmio_corrc5
		mov	r0,ZL
		mov	ZL,YL
		mov	YL,r0
libmio_corrc5:	cp	ZH,YH
		brcc	libmio_corrc6
		mov	r0,ZH
		mov	ZH,YH
		mov	YH,r0
libmio_corrc6:	ret		

;-------------------------------------------------------------------------------
; correct plot coordinates from YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
libmio_corrp:	cp	ZL,YL
		brcc	libmio_corrp1
		mov	r0,ZL
		mov	ZL,YL
		mov	YL,r0
libmio_corrp1:	cp	ZH,YH
		brcc	libmio_corrp2
		mov	r0,ZH
		mov	ZH,YH
		mov	YH,r0
libmio_corrp2:	ret		

;-------------------------------------------------------------------------------
; cbox YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
.ifdef use_libmio_cbox
libmio_cbox:	push	XL			;save registers
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		push	tempreg2
		rcall	libmio_corrc		;correct coords
libmio_cbox_2:	mov	XL,YL			;set start
		mov	XH,YH
libmio_cbox_3:	push	YL
		push	YH
		rcall	libmio_calc1
		st	Y,byte0
		ldi	tempreg2,LOW(libmio_cols*libmio_rows)
		add	YL,tempreg2
		ldi	tempreg2,HIGH(libmio_cols*libmio_rows)
		adc	YH,tempreg2
		lds	tempreg2,libmio_color
		st	Y,tempreg2
		pop	YH
		pop	YL
		cp	XL,ZL
		breq	libmio_cbox_4
		inc	XL
		rjmp	libmio_cbox_3
libmio_cbox_4:	cp	XH,ZH
		breq	libmio_cbox_5
		mov	XL,YL
		inc	XH
		rjmp	libmio_cbox_3
libmio_cbox_5:	pop	tempreg2
		pop	ZH
		pop	ZL
		pop	YH
		pop	YL
		pop	XH
		pop	XL
		ret				
.endif				

;-------------------------------------------------------------------------------
; ibox YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
.ifdef use_libmio_ibox
libmio_ibox:	push	XL			;save registers
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		push	tempreg1
		push	tempreg2
		rcall	libmio_corrc		;correct coords
libmio_ibox_2:	mov	XL,YL			;set start
		mov	XH,YH
libmio_ibox_3:	push	YL
		push	YH
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		ldi	tempreg2,LOW(libmio_cols*libmio_rows)
		add	YL,tempreg2
		ldi	tempreg2,HIGH(libmio_cols*libmio_rows)
		adc	YH,tempreg2
		add	YL,XL
		adc	YH,byte0		
		ldi	tempreg2,libmio_cols
		mul	XH,tempreg2
		add	YL,r0
		adc	YH,r1
		ld	tempreg2,Y
		mov	tempreg1,tempreg2
		andi	tempreg2,0xee
		andi	tempreg1,0x11
		swap	tempreg2
		or	tempreg2,tempreg1
		st	Y,tempreg2
		pop	YH
		pop	YL
		cp	XL,ZL
		breq	libmio_ibox_4
		inc	XL
		rjmp	libmio_ibox_3
libmio_ibox_4:	cp	XH,ZH
		breq	libmio_ibox_5
		mov	XL,YL
		inc	XH
		rjmp	libmio_ibox_3
libmio_ibox_5:	pop	tempreg2
		pop	tempreg1
		pop	ZH
		pop	ZL
		pop	YH
		pop	YL
		pop	XH
		pop	XL
		ret						
.endif				

;-------------------------------------------------------------------------------
; box YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
.ifdef use_libmio_box
libmio_box:	push	XL
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		push	tempreg2
		rcall	libmio_corrp
libmio_box_2:	mov	XL,YL
		mov	XH,YH
libmio_box_3:	rcall	libmio_plot
		inc	XL
		cp	XL,ZL
		brne	libmio_box_3
libmio_box_4:	rcall	libmio_plot
		inc	XH
		cp	ZH,XH
		brne	libmio_box_4
libmio_box_5:	rcall	libmio_plot
		dec	XL
		cp	XL,YL
		brne	libmio_box_5
libmio_box_6:	rcall	libmio_plot
		dec	XH
		cp	XH,YH
		brne	libmio_box_6
		pop	tempreg2
		pop	ZH
		pop	ZL
		pop	YH
		pop	YL
		pop	XH
		pop	XL
		ret				
.endif				

;-------------------------------------------------------------------------------
; fbox YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
.ifdef use_libmio_fbox
libmio_fbox:	push	XL
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		push	tempreg2
		rcall	libmio_corrp
libmio_fbox_2:	mov	XH,YH
libmio_fbox_3:	mov	XL,YL
libmio_fbox_4:	rcall	libmio_plot
		cp	XL,ZL
		breq	libmio_fbox_5
		inc	XL
		rjmp	libmio_fbox_4
libmio_fbox_5:	rcall	libmio_plot
		cp	ZH,XH
		breq	libmio_fbox_6
		inc	XH
		rjmp	libmio_fbox_3
.endif
		
libmio_fbox_6:	pop	tempreg2
libmio_fbox_7:	pop	ZH
		pop	ZL
		pop	YH
		pop	YL
		pop	XH
		pop	XL
		ret				

;-------------------------------------------------------------------------------
; windowbox YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
.ifdef use_libmio_wbox
libmio_wbox:	rcall	libmio_cbox		;clear box
		push	XL
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		push	tempreg1
		push	tempreg2
		rcall	libmio_corrc		;sort data
		mov	tempreg1,YL		;save X-pos
		mov	XH,YH
		mov	XL,YL
		rcall	libmio_calc1		;get address
		ldi	tempreg2,0xf2		;upper left corner
		st	Y+,tempreg2		;store
		ldi	tempreg2,0xf1		;hor line
libmio_wbox1:	inc	XL
		cp	XL,ZL
		breq	libmio_wbox2
		st	Y+,tempreg2		;store
		rjmp	libmio_wbox1		;loop
		
libmio_wbox2:	ldi	tempreg2,0xf3		;upper right corner
		st	Y+,tempreg2		;store
libmio_wbox3:	mov	XL,tempreg1		;restore X
		inc	XH			;next line
		cp	XH,ZH			;last line?
		breq	libmio_wbox4		;yes
		rcall	libmio_calc1		;get address
		ldi	tempreg2,0xf0		;vertical line
		st	Y+,tempreg2
		mov	XL,ZL
		rcall	libmio_calc1		;get address
		st	Y+,tempreg2
		rjmp	libmio_wbox3									
libmio_wbox4:	mov	XL,tempreg1
		rcall	libmio_calc1		;get address
		ldi	tempreg2,0xf4		;lower left corner
		st	Y+,tempreg2		;store
		ldi	tempreg2,0xf1		;hor line
libmio_wbox5:	inc	XL
		cp	XL,ZL
		breq	libmio_wbox6
		st	Y+,tempreg2		;store
		rjmp	libmio_wbox5		;loop
		
libmio_wbox6:	ldi	tempreg2,0xf5		;lower right corner
		st	Y+,tempreg2		;store
				
		pop	tempreg2
		pop	tempreg1
		pop	ZH
		pop	ZL
		pop	YH
		pop	YL
		pop	XH
		pop	XL
		ret				

.endif

;-------------------------------------------------------------------------------
; windowheadbox YL,YH to ZL,ZH
;-------------------------------------------------------------------------------
.ifdef use_libmio_whbox
libmio_whbox:	rcall	libmio_wbox		;draw normal windowbox
		push	tempreg2
		push	XL
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
	
		rcall	libmio_corrc		;sort data
		mov	XH,YH
		mov	XL,YL
		rcall	libmio_calc1		;get address
		ldi	tempreg2,0xfb		;left corner
		st	Y+,tempreg2		;store
		ldi	tempreg2,0xec		;header fields
libmio_whbox1:	inc	XL
		cp	XL,ZL
		breq	libmio_whbox2
		st	Y+,tempreg2		;store
		rjmp	libmio_whbox1		;loop
libmio_whbox2:	ldi	tempreg2,0xfc		;right corner
		st	Y+,tempreg2		;store

		pop	ZH			;restore coordinates
		pop	ZL
		pop	YH
		pop	YL
		push	YL
		push	YH
		push	ZL
		push	ZH

		mov	XH,YH
		mov	XL,YL
		inc	XH			;Y+1
		rcall	libmio_calc1		;get address
		ldi	tempreg2,0xfd		;left side
		st	Y+,tempreg2		;store
		ldi	tempreg2,0x20		;space
libmio_whbox3:	inc	XL
		cp	XL,ZL
		breq	libmio_whbox4
		st	Y+,tempreg2		;store
		rjmp	libmio_whbox3		;loop
libmio_whbox4:	ldi	tempreg2,0xfe		;right side
		st	Y+,tempreg2		;store

		pop	ZH			;restore coordinates
		pop	ZL
		pop	YH
		pop	YL
		push	YL
		push	YH
		push	ZL
		push	ZH

		mov	XH,YH
		mov	XL,YL
		inc	XH			;Y+1
		push	XL
		rcall	libmio_calc1		;get address
		ldi	XL,LOW(libmio_cols*libmio_rows)
		ldi	XH,HIGH(libmio_cols*libmio_rows)
		add	YL,XL
		adc	YH,XH
		lds	tempreg2,libmio_color	;get color
		swap	tempreg2
		andi	tempreg2,0xee		;only charset 0
		adiw	YL,0x01			;skip first
		pop	XL
libmio_whboxc3:	inc	XL
		cp	XL,ZL
		breq	libmio_whboxc4
		st	Y+,tempreg2		;store
		rjmp	libmio_whboxc3		;loop

libmio_whboxc4:	pop	ZH			;restore coordinates
		pop	ZL
		pop	YH
		pop	YL
		push	YL
		push	YH
		push	ZL
		push	ZH

		mov	XH,YH
		mov	XL,YL
		subi	XH,0xfe			;Y+2
		rcall	libmio_calc1		;get address
		ldi	tempreg2,0xee		;left side
		st	Y+,tempreg2		;store
		ldi	tempreg2,0xed		;fill
libmio_whbox5:	inc	XL
		cp	XL,ZL
		breq	libmio_whbox6
		st	Y+,tempreg2		;store
		rjmp	libmio_whbox5		;loop
libmio_whbox6:	ldi	tempreg2,0xef		;right side
		st	Y+,tempreg2		;store

		pop	ZH
		pop	ZL
		pop	YH
		pop	YL
		pop	XH
		pop	XL
		pop	tempreg2
		ret				
.endif

;-------------------------------------------------------------------------------
; draw line from YL,YH to ZL,ZH
; this algorithm is a little bit quick and dirty, but it works...
;-------------------------------------------------------------------------------
.ifdef use_libmio_draw
libmio_draw:	push	r16
		push	r17
		push	r18
		push	r19
		push	XL
		push	XH
		push	YL
		push	YH		
		movw	XL,YL			;copy start pointer
		ldi	r16,0xff		;set sgn(dx) negative
		sub	YL,ZL			;x1-x2
		brcc	libmio_draw01		;branch if x1<x2
		neg	r16			;invert sgn(dx)
		neg	YL			;invert dx
libmio_draw01:	ldi	r17,0xff		;set sgn(dy) negative
		sub	YH,ZH			;-dy
		brcc	libmio_draw02		;branch if y1<y2
		neg	r17			;invert sgn(dy)
		neg	YH			;invert dy
libmio_draw02:	cp	YL,YH			;|dx|-|dy|
		brcs	libmio_draw6		;branch if |dy|>|dx|
		rjmp	libmio_draw3

		

;|dx|>|dy|		
libmio_draw3:	mov	r18,YL			;|dx|
		lsr	r18			;/2
libmio_draw4:	rcall	libmio_plot		;plot pixel
		cp	XL,ZL			;the endpoint?
		breq	libmio_draw_e		;yes
		add	XL,r16			;X-step
		sub	r18,YH			;-|dy|
		brcc	libmio_draw4		;no overflow
		add	XH,r17			;y-step
		add	r18,YL
		rjmp	libmio_draw4			

;|dy|>|dx|
libmio_draw6:	mov	r18,YH			;|dy|
		lsr	r18			;/2
libmio_draw7:	rcall	libmio_plot		;plot pixel
		cp	XH,ZH			;the endpoint?
		breq	libmio_draw_e
		add	XH,r17			;y-step
		sub	r18,YL
		brcc	libmio_draw7			
		add	XL,r16			;x-step
		add	r18,YH
		rjmp	libmio_draw7			

libmio_draw_e:	pop	YH
		pop	YL
		pop	XH
		pop	XL
		pop	r19
		pop	r18
		pop	r17
		pop	r16
		ret		


.endif

;-------------------------------------------------------------------------------
; scroll XL lines 1 character left
;-------------------------------------------------------------------------------
.ifdef use_libmio_lscroll
libmio_lsc:	push	XL
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)	
		clr	ZL
libmio_lsc_01:	ldi	ZH,libmio_cols-1
libmio_lsc_02:	ldd	XH,Y+1
		st	Y+,XH
		dec	ZH
		brne	libmio_lsc_02
		st	Y+,byte0
		inc	ZL
		cp	ZL,XL
		breq	libmio_lsc_03
		cpi	ZL,libmio_rows
		brne	libmio_lsc_01
libmio_lsc_03:	ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	ZH,LOW(libmio_cols*libmio_rows)
		add	YL,ZH
		ldi	ZH,HIGH(libmio_cols*libmio_rows)
		adc	YH,ZH
		clr	ZL
libmio_lsc_11:	ldi	ZH,libmio_cols-1
libmio_lsc_12:	ldd	XH,Y+1
		st	Y+,XH
		dec	ZH
		brne	libmio_lsc_12
		st	Y+,byte0
		inc	ZL
		cp	ZL,XL
		breq	libmio_lsc_13
		cpi	ZL,libmio_rows
		brne	libmio_lsc_11
libmio_lsc_13:	rjmp	libmio_fbox_7
.endif


;-------------------------------------------------------------------------------
; scroll XL lines 1/2 character left (1 plot pixel)
;-------------------------------------------------------------------------------
.ifdef use_libmio_lpscroll
libmio_lps:	push	XL
		push	XH
		push	YL
		push	YH
		push	ZL
		push	ZH
		push	r16
		push	r17
		ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)	
		clr	ZL
libmio_lps_01:	ld	r16,Y
		ldi	ZH,libmio_cols-1
libmio_lps_02:	ldd	r17,Y+1
		bst	r16,1
		bld	r16,0
		bst	r17,0
		bld	r16,1
		bst	r16,3
		bld	r16,2
		bst	r17,2
		bld	r16,3
		andi	r16,0x0f
		st	Y+,r16
		mov	r16,r17
		dec	ZH
		brne	libmio_lps_02
		bst	r17,1
		bld	r17,0
		bst	r17,3
		bld	r17,2
		andi	r17,0x05
		st	Y+,r17
		inc	ZL
		cp	ZL,XL
		breq	libmio_lps_03
		cpi	ZL,libmio_rows
		brne	libmio_lps_01
libmio_lps_03:	pop	r17
		pop	r16
		rjmp	libmio_fbox_7
.endif


.ifdef use_libmio_sprites
;-------------------------------------------------------------------------------
; sprite definitions
; +00  1 byte type
; +01  4 bytes chars
; +05  4 bytes colours
; +09  4 bytes stored chars
; +0d  4 bytes stored colours
; +11  2 bytes coord
; +13  1 byte status
;-------------------------------------------------------------------------------
;-------------------------------------------------------------------------------
; calc memory address of sprite ctrl to Z
;-------------------------------------------------------------------------------
libmio_scalc:	push	r16
		ldi	r16,0x14	;20 bytes/sprite
		mul	r16,ctrl
		ldi	ZL,LOW(libmio_sprites)
		ldi	ZH,HIGH(libmio_sprites)
		add	ZL,r0
		adc	ZH,r1
		pop	r16
		ret

;-------------------------------------------------------------------------------
; set sprite,  XH=nr
;-------------------------------------------------------------------------------
libmio_sdef:	push	ZL		;save pointer
		push	ZH
		push	r16
		andi	ctrl,0x07
		rcall	libmio_scalc	;sprite address
		std	Z+0x13,byte0	;clear status
libmio_sdef1:	pop	r16
		pop	ZH
		pop	ZL
		ret


;-------------------------------------------------------------------------------
; test sprite, ctrl=sprite no
;-------------------------------------------------------------------------------
libmio_stest:	rcall	libmio_scalc		;calc sprite address
libmio_stest1:	ldd	XL,Z+0x11		;old X
		ldd	XH,Z+0x12		;old Y
		ld	tempreg1,Z		;get mode
		adiw	ZL,9			;set pointer
		rcall	libmio_ssave		;save first char	
		cpi	tempreg1,0		;1 char
		breq	libmio_steste		;ok
		
		cpi	tempreg1,4		;hor?
		brcc	libmio_stest3		;no
libmio_stest2:	inc	XL			;X+1
		rcall	libmio_ssave		;save next char
		dec	tempreg1
		brne	libmio_stest2
		ret				;thats all
		
libmio_stest3:	breq	libmio_stest5
		subi	tempreg1,4
libmio_stest4:	inc	XH			;Y+1
		rcall	libmio_ssave		;save next char
		dec	tempreg1
		brne	libmio_stest4
		ret
						;thats all
libmio_stest5:	inc	XL			;X+1
		rcall	libmio_ssave		;save next char
		dec	XL
		inc	XH
		rcall	libmio_ssave		;save next char
		inc	XL
		rcall	libmio_ssave		;save next char
libmio_steste:	ret				;thats all
				

;-------------------------------------------------------------------------------
; save background char,  XL/XH=position (Z)=char t=collision
;-------------------------------------------------------------------------------
libmio_ssave:	push	r16
		cpi	XL,libmio_cols		;visible in X range?
		brcc	libmio_ssavee		;no
		cpi	XH,libmio_rows		;visible in Y range?
		brcc	libmio_ssavee		;no
		rcall	libmio_calc1		;get char position
		ld	r16,Y			;get char in vram
		st	Z,r16			;store in buffer
		andi 	r16,0x0f		;n0 characters give no collision
		breq	libmio_ssave1		;
		set				;set collision flag
		
libmio_ssave1:	ldi	r16,LOW(libmio_cols*libmio_rows)
		add	YL,r16
		ldi	r16,HIGH(libmio_cols*libmio_rows)
		adc	YH,r16
		ld	r16,Y			;get colour from vram
		std	Z+4,r16			;store in buffer

libmio_ssavee:	pop	r16
		adiw	ZL,1			;increment pointer
		ret				;thats all

;-------------------------------------------------------------------------------
; recover background char,  XL/XH=position (Z)=char store 9..12
;-------------------------------------------------------------------------------
libmio_srec:	push	r16
		cpi	XL,libmio_cols		;visible in X range?
		brcc	libmio_srece		;no
		cpi	XH,libmio_rows		;visible in Y range?
		brcc	libmio_srece		;no
		rcall	libmio_calc1		;get char location
		ld	r16,Z			;get char in buffer
		st	Y,r16			;store in vram
		ldi	r16,LOW(libmio_cols*libmio_rows)
		add	YL,r16
		ldi	r16,HIGH(libmio_cols*libmio_rows)
		adc	YH,r16
		ldd	r16,Z+4			;get char in buffer
		st	Y,r16			;store in vram
libmio_srece:	pop	r16
		adiw	ZL,1			;next pos
		ret				;thats all

;-------------------------------------------------------------------------------
; restore background under sprite, ctrl=sprite no
;-------------------------------------------------------------------------------
libmio_sres:	rcall	libmio_scalc		;calc sprite address
		ldd	tempreg1,Z+0x13		;status
		cpi	tempreg1,1		;visible
		breq	libmio_sres1		;yes
		ret				;nothing to do
		
libmio_sres1:	ldd	XL,Z+0x11		;old X
		ldd	XH,Z+0x12		;old Y
		ld	tempreg1,Z		;get mode
		adiw	ZL,9			;set pointer to first char
		rcall	libmio_srec		;restore first char
			
		cpi	tempreg1,0		;1 char
		breq	libmio_srese		;ok
		
		cpi	tempreg1,4		;hor?
		brcc	libmio_sres3		;no
libmio_sres2:	inc	XL			;X+1
		rcall	libmio_srec		;recover next char
		dec	tempreg1
		brne	libmio_sres2
		ret				;thats all
		
libmio_sres3:	breq	libmio_sres5
		subi	tempreg1,4
libmio_sres4:	inc	XH			;Y+1
		rcall	libmio_srec		;recover next char
		dec	tempreg1
		brne	libmio_sres4
		ret				;thats all
		
libmio_sres5:	inc	XL			;X+1
		rcall	libmio_srec		;recover next char
		dec	XL
		inc	XH
		rcall	libmio_srec		;recover next char
		inc	XL
		rcall	libmio_srec		;recover next char
libmio_srese:	ret				;thats all

;-------------------------------------------------------------------------------
; draw sprite char,  XL/XH=position (Z)=char
;-------------------------------------------------------------------------------
libmio_sdrawc:	push	r16
		cpi	XL,libmio_cols		;visible in X range?
		brcc	libmio_sdrawce		;no
		cpi	XH,libmio_rows		;visible in Y range?
		brcc	libmio_sdrawce		;no
		rcall	libmio_calc1		;get char position
		ld	r16,Z			;get char
		st	Y,r16			;write to vram
		ldi	r16,LOW(libmio_cols*libmio_rows)
		add	YL,r16
		ldi	r16,HIGH(libmio_cols*libmio_rows)
		adc	YH,r16
		ldd	r16,Z+4			;get colour
		push	r17
		ld	r17,Y
		andi	r17,0xe1
		lsl	r16
		andi	r16,0x1e
		or	r16,r17
		pop	r17
		st	Y,r16			;write colour
libmio_sdrawce:	pop	r16
		adiw	ZL,1			;increment pointer1
		ret				;thats all

;-------------------------------------------------------------------------------
; draw sprite, ctrl=sprite no
;-------------------------------------------------------------------------------
libmio_sdraw:	rcall	libmio_scalc		;calc sprite address
libmio_sdraw1:	ldd	XL,Z+0x11		;old X
		ldd	XH,Z+0x12		;old Y
		ld	tempreg1,Z		;get mode
		adiw	ZL,1			;set pointer
		rcall	libmio_sdrawc		;draw first char	
		cpi	tempreg1,0		;1 char
		breq	libmio_sdrawe		;ok
		cpi	tempreg1,4		;hor?
		brcc	libmio_sdraw3		;no
libmio_sdraw2:	inc	XL			;X+1
		rcall	libmio_sdrawc		;draw next char
		dec	tempreg1
		brne	libmio_sdraw2
		ret				;thats all
		
libmio_sdraw3:	breq	libmio_sdraw5
		subi	tempreg1,4
libmio_sdraw4:	inc	XH			;Y+1
		rcall	libmio_sdrawc		;draw next char
		dec	tempreg1
		brne	libmio_sdraw4
		ret				;thats all
		
libmio_sdraw5:	inc	XL			;X+1
		rcall	libmio_sdrawc		;draw next char
		dec	XL
		inc	XH
		rcall	libmio_sdrawc		;draw next char
		inc	XL
		rcall	libmio_sdrawc		;draw next char
libmio_sdrawe:	ret				;thats all
				

;-------------------------------------------------------------------------------
; show sprite,  ctrl=nr  XH=y XL=x
;-------------------------------------------------------------------------------
libmio_sshow:	push	ZL			;save registers
		push	ZH
		push	r16
		push	YL
		push	YH
		push	XL
		push	XH
		andi	ctrl,0x07
		rcall	libmio_sres		;restore old if necessary
		pop	XH
		pop	XL
		rcall	libmio_scalc		;get sprite address
		std	Z+0x11,XL		;new X
		std	Z+0x12,XH		;new Y
		clt				;clear col flag
		rcall	libmio_stest		;test for collisions
		brtc	libmio_sshow1		;no
		rcall	libmio_scalc		;get sprite address
		ldi	r16,2			;set collision
		rjmp	libmio_sshow2
		
libmio_sshow1:	rcall	libmio_sdraw		;draw new sprite
		rcall	libmio_scalc		;get sprite address
		ldi	r16,1			;set visible flag
libmio_sshow2:	std	Z+0x13,r16
		pop	YH
		pop	YL
		rjmp	libmio_sdef1

;-------------------------------------------------------------------------------
; hide sprite,  ctrl=nr
;-------------------------------------------------------------------------------
libmio_shide:	push	ZL			;save registers
		push	ZH
		push	r16
		push	YL
		push	YH
		push	XL
		push	XH
		andi	ctrl,0x07
		rcall	libmio_sres		;restore old if necessary
		pop	XH
		pop	XL
		clr	r16			;clear visible and collision
		rjmp	libmio_sshow2

.endif
				
;-------------------------------------------------------------------------------
; wait for the end of visible frame
;-------------------------------------------------------------------------------
.ifdef use_libmio_sync
libmio_sync:	push	XL
libmio_sync_1:	mov	XL,vline_l
		cpi	XL,libmio_pixel_rows+1
		breq	libmio_sync_1
libmio_sync_2:	mov	XL,vline_l
		cpi	XL,libmio_pixel_rows+1
		brne	libmio_sync_2
		pop	XL
		ret		    
.endif

;-------------------------------------------------------------------------------
; set palette to YH,YL,ZH,ZL
;-------------------------------------------------------------------------------
.ifdef use_libmio_setpal
libmio_setpal:	swap	YH
		lsl	YH
		andi	YH,0xe0
		sts	libmio_palette+0,YH
		swap	YL
		lsl	YL
		andi	YL,0xe0
		sts	libmio_palette+1,YL

		swap	ZH
		lsl	ZH
		andi	ZH,0xe0
		sts	libmio_palette+2,ZH
		swap	ZL
		lsl	ZL
		andi	ZL,0xe0
		sts	libmio_palette+3,ZL
		ret
.endif

;-------------------------------------------------------------------------------
; play a note (XL=note 0..63 0xff=>noise)
;-------------------------------------------------------------------------------
.ifdef use_libmio_note
libmio_note:	push	ZL
		push	ZH
		push	XL
		sts	libmio_note_val,XL	;write note value
		andi	XL,0x3f			;64 notes
		lsl	XL
		ldi	ZH,HIGH(libmio_ntable*2)	;1 start envtable high
		ldi	ZL,LOW(libmio_ntable*2)	;1 start envtable low
		add	ZL,XL			;offsettable
		adc	ZH,byte0		;
		lpm	XL,Z+			;get low offset
		sts	libmio_woff_l,XL	;write
		lpm	XL,Z+			;get high offset
		sts	libmio_woff_h,XL	;write
		sts	libmio_env,byte0	;start
		sts	libmio_wpos_l,byte0
		sts	libmio_wpos_h,byte0
		pop	XL 
		pop	ZH
		pop	ZL
		ret		    
.endif

;------------------------------------------------------------------------------
;wait for a key
;------------------------------------------------------------------------------
.ifdef use_libmio_waitkey
libmio_waitkey:	sts	libmio_lastkey,byte0
libmio_wkey1:	lds	tempreg1,libmio_lastkey
		cpi	tempreg1,0
		breq	libmio_wkey1

.ifdef use_libmio_screenshot		
		rjmp	libmio_scrs
.else
		ret
.endif				
.endif

;------------------------------------------------------------------------------
;wait for no key
;------------------------------------------------------------------------------
libmio_nokey:	lds	tempreg1,libmio_keycode
		cpi	tempreg1,0
		brne	libmio_nokey
		ret

		
;------------------------------------------------------------------------------
;blink cursor and wait for key (return x,y in X ; adr in Y ; key in tempreg1)
;------------------------------------------------------------------------------
.ifdef use_libmio_cursor
libmio_cursor:	rcall	libmio_calc		;calculate address
		ldi	XL,LOW(690)
		add	YL,XL
		ldi	XL,HIGH(690)
		adc	YH,XL
		
libmio_cur1:	ld	XL,Y			;get original color
		mov	XH,XL			;copy
		push	XL
		andi	XL,0x11
		andi	XH,0xee
		swap	XH
		or	XH,XL
		pop	XL
		sts	libmio_lastkey,byte0	;clear key
libmio_cur_1:	lds	tempreg1,libmio_lastkey
		cpi	tempreg1,0
		brne	libmio_cur_e

		lds	tempreg1,libmio_frames
		cpi	tempreg1,27		;ca. half of fps
		brcc	libmio_cur_2
		st	Y,XL
		rjmp	libmio_cur_1
		
libmio_cur_2:	st	Y,XH
		rjmp	libmio_cur_1						
		
libmio_cur_e:	st	Y,XL			;put back original char
		
.ifdef use_libmio_screenshot		
		rcall	libmio_scrs
.endif				
		rjmp	libmio_calc
.endif			

;------------------------------------------------------------------------------
;blink cursor and wait for key (return x,y in X ; adr in Y ; key in tempreg1)
;------------------------------------------------------------------------------
.ifdef use_libmio_ecursor
libmio_ecursor:	push	r19
		ldi	r19,0xdd		;dummy
		rcall	libmio_ecalc
		sts	libmio_lastkey,byte0	;clear key
libmio_ecur_1:	lds	tempreg1,libmio_kbdstate
		andi	tempreg1,0x04		;left ctrl
		cp	tempreg1,r19
		breq	libmio_ecur_c
		cpi	tempreg1,0x00		;no
		brne	libmio_ecur_b
		call	editor_shline
		mov	r19,tempreg1
		rjmp	libmio_ecur_c
libmio_ecur_b:	call	editor_chline
		mov	r19,tempreg1
libmio_ecur_c:	lds	tempreg1,libmio_lastkey
		cpi	tempreg1,0
		brne	libmio_ecur_e
		lds	tempreg1,libmio_frames
		cpi	tempreg1,27		;ca. half of fps
		brcc	libmio_ecur_2
		ld	tempreg1,Y		;get char
		andi	tempreg1,0x7f		;set normal
		st	Y,tempreg1
		rjmp	libmio_ecur_1
libmio_ecur_2:	ld	tempreg1,Y		;get char
		ori	tempreg1,0x80		;set inverse
		st	Y,tempreg1
		rjmp	libmio_ecur_1						
		
libmio_ecur_e:	ld	tempreg2,Y		;get char
		andi	tempreg2,0x7f		;set normal
		st	Y,tempreg2
		pop	r19
		ret		
.endif

;------------------------------------------------------------------------------
;blink cursor and wait for key (return x,y in X ; adr in Y ; key in tempreg1)
;------------------------------------------------------------------------------
.ifdef use_libmio_ncursor
libmio_ncursor:	ldi	YH,HIGH(libmio_vram+11)	;home position
		ldi	YL,LOW(libmio_vram+11)	;home position
		add	YL,XL			;position+x
		adc	YH,byte0		
		sts	libmio_lastkey,byte0	;clear key
libmio_ncur_1:	lds	tempreg1,libmio_lastkey
		cpi	tempreg1,0
		brne	libmio_ncur_e
		lds	tempreg1,libmio_frames
		cpi	tempreg1,27		;ca. half of fps
		brcc	libmio_ncur_2
		ld	tempreg1,Y		;get char
		andi	tempreg1,0x7f		;set normal
		st	Y,tempreg1
		rjmp	libmio_ncur_1
libmio_ncur_2:	ld	tempreg1,Y		;get char
		ori	tempreg1,0x80		;set inverse
		st	Y,tempreg1
		rjmp	libmio_ncur_1						
		
libmio_ncur_e:	ld	tempreg2,Y		;get char
		andi	tempreg2,0x7f		;set normal
		st	Y,tempreg2
		ret		
.endif

;------------------------------------------------------------------------------
;calculate char address to Y
;------------------------------------------------------------------------------
libmio_calc:	mov	XL,libmio_cur_x		;X
		mov	XH,libmio_cur_y		;Y
libmio_calc1:	ldi	YH,libmio_cols		;chars per line
		mul	XH,YH			;store
		ldi	YH,HIGH(libmio_vram)	;home position
		ldi	YL,LOW(libmio_vram)	;home position
		add	YL,XL			;position+x
		adc	YH,byte0		
		add	YL,r0			;position+(cols*y)
		adc	YH,r1
		ret

;------------------------------------------------------------------------------
;calculate attributes address to Y
;------------------------------------------------------------------------------
libmio_calc2:	rcall	libmio_calc1
		push	r16
		ldi	r16,LOW(libmio_cols*libmio_rows)
		add	YL,r16
		ldi	r16,HIGH(libmio_cols*libmio_rows)
		adc	YH,r16
		pop	r16
		ret
		
;------------------------------------------------------------------------------
;calculate ecursor address to Y
;------------------------------------------------------------------------------
.ifdef use_libmio_editor_mode		
libmio_ecalc:	ldi	YH,libmio_ecols		;chars per line
		mul	XH,YH			;store
		ldi	YH,HIGH(libmio_vram+3*libmio_ecols+3)	;home position
		ldi	YL,LOW(libmio_vram+3*libmio_ecols+3)	;home position
		add	YL,XL			;position+x
		adc	YH,byte0		
		add	YL,r0			;position+(cols*y)
		adc	YH,r1
		ret
.endif

;------------------------------------------------------------------------------
;get char on cursor address
;------------------------------------------------------------------------------
.ifdef use_libmio_gchar
libmio_gchar:	rcall	libmio_calc
		ld	tempreg1,Y
		ret
.endif		

;------------------------------------------------------------------------------
;set cursor to down position
;------------------------------------------------------------------------------
.ifdef use_libmio_vcur
libmio_vcur:	mov	XL,libmio_cur_x		;X
		mov	XH,libmio_cur_y		;Y
		dec	XL
		inc	XH
		rjmp	libmio_gxy		;set new cursor position
.endif		
		
;------------------------------------------------------------------------------
;put tempreg1 to printer output
;------------------------------------------------------------------------------
libmio_ppar:	push	XH
		ldi	XH,0xff
		out	DDRA,XH			;set port A4-7 as outputs
		nop
libmio_ppar_1:	sbic	PINB,2			;busy
		rjmp	libmio_ppar_1		;yes
		out	PORTA,tempreg1		;data
		ldi	XH,100			;pause
libmio_ppar_2:	dec	XH
		brne	libmio_ppar_2
		cbi	PORTB,1			;clear strobe (active)
		ldi	XH,100			;pause
libmio_ppar_3:	dec	XH
		brne	libmio_ppar_3
		sbi	PORTB,1			;set strobe (inactive)
		pop	XH
		ret				;thats all


;------------------------------------------------------------------------------
;output screenshot
;------------------------------------------------------------------------------
.ifdef use_libmio_screenshot
libmio_scrs_1:	pop	tempreg1
		ret
libmio_scrs:	push	tempreg1		;
		lds	tempreg1,libmio_kflags
		sbrs	tempreg1,1		;screenshot-flag
		rjmp	libmio_scrs_1
		andi	tempreg1,0xfd		;clear screenshot-flag
		sts	libmio_kflags,tempreg1	;write back
		push	ZH			;save Z-register
		push	ZL			;save Z-register
		push	YH			;save Y-register
		push	YL			;save Y-register
		lds	tempreg1,libmio_vidmode	;video mode
		cpi	tempreg1,0x00		;text?
		brne	libmio_scrs_01		;no
		lds	tempreg1,libmio_config	;editor/normal
		sbrs	tempreg1,7		;check bit 7
		rjmp	libmio_scrs_2		;normal
		rjmp	libmio_scrs_5		;editor
libmio_scrs_01:	push	tempreg1
		ldi	tempreg1,'G'		;identify
		rcall	libmio_pser
		pop	tempreg1		;reload
		push	tempreg1
		subi	tempreg1,0xd0		;+0x30
		rcall	libmio_pser
		pop	tempreg1		;restore
		cpi	tempreg1,0x01		;mode 1
		brne	libmio_scrs_02		
		rjmp	libmio_scrg1		;screenshot mode 1
libmio_scrs_02:	cpi	tempreg1,0x02		;mode 2
		brne	libmio_scrs_03		
		rjmp	libmio_scrg2		;screenshot mode 2
libmio_scrs_03:	rjmp	libmio_scrg3	

		
libmio_scrs_2:	ldi	tempreg1,'T'		;identify
		call	libmio_pser
		ldi	tempreg1,0x0a
		call	libmio_pser

		ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	ZH,libmio_rows*2	;number of charlines
libmio_scrs_3:	ldi	ZL,libmio_cols		;number of chars per line
libmio_scrs_4:	ld	tempreg1,Y
		swap	tempreg1		;high nibble first
		rcall	libmio_scrs_c
		ld	tempreg1,Y+
		rcall	libmio_scrs_c
		dec	ZL
		brne	libmio_scrs_4
		ldi	tempreg1,0x0a
		call	libmio_pser
		dec	ZH
		brne	libmio_scrs_3
		pop	YL
		pop	YH
		pop	ZL
		pop	ZH
		pop	tempreg1
		clr	tempreg1		;kill character		
		ret

libmio_scrs_5:	ldi	tempreg1,'E'		;identify
		call	libmio_pser		;out
		lds	tempreg1,libmio_estatus	;color of status line
		rcall	libmio_scrs_cc		;out
		ldi	tempreg1,0x0a
		call	libmio_pser

		ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	ZH,0x03			;number of fix charlines
libmio_scrs_6:	ldi	ZL,libmio_ecols		;number of chars per line
libmio_scrs_7:	ld	tempreg1,Y
		swap	tempreg1		;high nibble first
		rcall	libmio_scrs_c
		ld	tempreg1,Y+
		rcall	libmio_scrs_c
		dec	ZL
		brne	libmio_scrs_7
		ldi	tempreg1,0x0a
		call	libmio_pser
		dec	ZH
		brne	libmio_scrs_6
		ldi	ZL,libmio_ecols		;calc startadr of listing 
		lds	ZH,libmio_edstart
		mul	ZL,ZH
		add	YL,r0
		adc	YH,r1
		ldi	ZH,0x10			;number of edit charlines
libmio_scrs_8:	ldi	ZL,libmio_ecols		;number of chars per line
libmio_scrs_9:	ld	tempreg1,Y
		andi	tempreg1,0x7f		;mask invert bit
		swap	tempreg1		;high nibble first
		rcall	libmio_scrs_c
		ld	tempreg1,Y+
		andi	tempreg1,0x7f		;mask invert bit
		rcall	libmio_scrs_c
		dec	ZL
		brne	libmio_scrs_9
		ldi	tempreg1,0x0a
		call	libmio_pser
		dec	ZH
		brne	libmio_scrs_8
		pop	YL
		pop	YH
		pop	ZL
		pop	ZH
		pop	tempreg1
		clr	tempreg1		;kill character		
		ret

;graphics mode 1	
libmio_scrg1:	lds	tempreg1,0x100		;palette enty 0
		rcall	libmio_scrs_cc		;out
		lds	tempreg1,0x101		;palette enty 1
		rcall	libmio_scrs_cc		;out
		ldi	tempreg1,0x0a
		rcall	libmio_pser

libmio_scrg1_1:	ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	ZH,116			;number of pixlines
libmio_scrg1_2:	ldi	ZL,21			;number of chars per line
libmio_scrg1_3:	ld	tempreg1,Y
		swap	tempreg1		;high nibble first
		rcall	libmio_scrs_c
		ld	tempreg1,Y+
		rcall	libmio_scrs_c
		dec	ZL
		brne	libmio_scrg1_3
		ldi	tempreg1,0x0a
		call	libmio_pser
		dec	ZH
		brne	libmio_scrg1_2
		pop	YL
		pop	YH
		pop	ZL
		pop	ZH
		pop	tempreg1
		clr	tempreg1		;kill character		
		ret

;graphics mode 2	
libmio_scrg2:	lds	tempreg1,0x100		;palette enty 0
		rcall	libmio_scrs_cc		;out
		lds	tempreg1,0x101		;palette enty 1
		rcall	libmio_scrs_cc		;out
		lds	tempreg1,0x102		;palette enty 2
		rcall	libmio_scrs_cc		;out
		lds	tempreg1,0x103		;palette enty 3
		rcall	libmio_scrs_cc		;out
		ldi	tempreg1,0x0a
		rcall	libmio_pser

libmio_scrg2_1:	ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	ZH,84			;number of pixlines
libmio_scrg2_2:	ldi	ZL,30			;number of chars per line
libmio_scrg2_3:	ld	tempreg1,Y
		swap	tempreg1		;high nibble first
		rcall	libmio_scrs_c
		ld	tempreg1,Y+
		rcall	libmio_scrs_c
		dec	ZL
		brne	libmio_scrg2_3
		ldi	tempreg1,0x0a
		call	libmio_pser
		dec	ZH
		brne	libmio_scrg2_2
		pop	YL
		pop	YH
		pop	ZL
		pop	ZH
		pop	tempreg1
		clr	tempreg1		;kill character		
		ret

;graphics mode 3	
libmio_scrg3:	ldi	tempreg1,0x0a
		rcall	libmio_pser

libmio_scrg3_1:	ldi	YL,LOW(libmio_vram)
		ldi	YH,HIGH(libmio_vram)
		ldi	ZH,58			;number of pixlines
libmio_scrg3_2:	ldi	ZL,42			;number of chars per line
libmio_scrg3_3:	ld	tempreg1,Y
		swap	tempreg1		;high nibble first
		rcall	libmio_scrs_c
		ld	tempreg1,Y+
		rcall	libmio_scrs_c
		dec	ZL
		brne	libmio_scrg3_3
		ldi	tempreg1,0x0a
		call	libmio_pser
		dec	ZH
		brne	libmio_scrg3_2
		pop	YL
		pop	YH
		pop	ZL
		pop	ZH
		pop	tempreg1
		clr	tempreg1		;kill character		
		ret


libmio_scrs_cc:	swap	tempreg1		;
		lsr	tempreg1
		andi	tempreg1,0x07		;mask bits
		
libmio_scrs_c:	andi	tempreg1,0x0f		;mask bits
		subi	tempreg1,0xd0		;+30
		cpi	tempreg1,0x3a		;>9
		brcs	libmio_scrs_c1		;jump, if not
		subi	tempreg1,0xf9		;+6
libmio_scrs_c1:	jmp	libmio_pser		

.endif

;------------------------------------------------------------------------------
;enable/disable chargepump
;------------------------------------------------------------------------------
.ifdef use_libmio_chp
libmio_chp:	push	XH
		ldi	XH,0x02			;disable output
		cpi	XL,0			;is zero
		breq	libmio_chp1
		ori	XH,0x40			;enable output
libmio_chp1:	out	TCCR0A,XH
		pop	XH
		ret
.endif


;------------------------------------------------------------------------------
;put tempreg1 to serial output
;------------------------------------------------------------------------------
libmio_pser:	push	XH
libmio_pser_1:	lds	XH,libmio_tcnt		;counter
		cpi	XH,0x00			;ready?
		brne	libmio_pser_1		;loop
		sts	libmio_tdat,tempreg1	;store data
		dec	XH			;prepare 0xff
		sts	libmio_tcnt,XH		;store start
		pop	XH
		ret				;thats all

;------------------------------------------------------------------------------
;get tempreg1 from serial input
;------------------------------------------------------------------------------
libmio_gser:	push	XH
libmio_gser_1:	lds	XH,libmio_rcnt		;counter
		cpi	XH,0x00			;ready?
		brne	libmio_gser_1		;loop
		dec	XH			;prepare 0xff
		sts	libmio_rcnt,XH		;store start
libmio_gser_2:	lds	XH,libmio_rcnt		;counter
		cpi	XH,0x00			;ready?
		brne	libmio_gser_2		;loop
		lds	tempreg1,libmio_rdat	;get char
libmio_pconv1:	pop	XH
		ret				;thats all

;------------------------------------------------------------------------------
;get tempreg1 from serial input with break
;------------------------------------------------------------------------------
libmio_gserb:	push	XH
libmio_gserb_1:	lds	XH,libmio_rcnt		;counter
		cpi	XH,0x00			;ready?
		brne	libmio_gserb_1		;loop
		dec	XH			;prepare 0xff
		sts	libmio_rcnt,XH		;store start
libmio_gserb_2:	lds	tempreg1,libmio_keycode	;get keycode
		cpi	tempreg1,0xed		;ESC
		breq	libmio_gserb_4		;end if ESC key is pressed
		lds	XH,libmio_rcnt		;counter
		cpi	XH,0x00			;ready?
		brne	libmio_gserb_2		;loop
		lds	tempreg1,libmio_rdat	;get char
libmio_gserb_3:	pop	XH
		ret				;thats all
libmio_gserb_4:	ldi	tempreg1,0x1b		;ESC
		pop	XH
		ret				;thats all


.include 	"libmio/bcopy.asm"
.include 	"libmio/circle.asm"
.include	"libmio/alert.asm"
					