;################################################################################
;#										#
;# avr-chipbasic2 - single chip basic computer with ATmega644			#
;# EEPROM/Flash memory interface						#
;# copyright (c) 2006/2007 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;#										#
;# This program is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU General Public License			#
;# as published by the Free Software Foundation; either version 2		#
;# of the License, or (at your option) any later version.			#
;#										#
;# This program is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

;------------------------------------------------------------------------------
; set kbdtype to eeprom
;------------------------------------------------------------------------------
mem_setktype:	lds	tempreg1,libmio_ksel	;set number
		ldi	YL,LOW(2040)		;default -> config 
		ldi	YH,HIGH(2040)
		libeep_write
		ret
	
;------------------------------------------------------------------------------
; get kbdtype
;------------------------------------------------------------------------------
mem_getktype:	ldi	YL,LOW(2040)		;default -> config 
		ldi	YH,HIGH(2040)
		libeep_read			;get value
		andi	tempreg1,0x03
		cpi	tempreg1,0x03
		brne	mem_getktype1
		clr	tempreg1
mem_getktype1:	sts	libmio_ksel,tempreg1
		ret		

;------------------------------------------------------------------------------
; set memory pointer to prog tempreg1
;------------------------------------------------------------------------------
mem_setprog:	push	ZH
		push	ZL	
		push	YH
		push	YL
		push	XL
		rcall	mem_progadr_0		;get ptr in Z
		sts	libmio_ram+44,ZL	;program base L
		sts	libmio_ram+45,ZH	;program base H
		pop	XL
		pop	YL
		pop	YH
		pop	ZL
		pop	ZH
		ret


mem_setthis:	ldi	ZL,LOW(bas_programs*2)	;base address
		ldi	ZH,HIGH(bas_programs*2)
		ldi	YL,LOW(3072)		;bytes/program
		ldi	YH,HIGH(3072)
		mov	XL,tempreg1
mem_setprog1:	cpi	XL,0
		breq	mem_setprog2		
		add	ZL,YL
		adc	ZH,YH
		dec	XL
		rjmp	mem_setprog1
mem_setprog2:	ret
						
;------------------------------------------------------------------------------
; read line (ctrl) from program memory to line buffer
;------------------------------------------------------------------------------
mem_readline:	lds	ZL,libmio_ram+44	;base address
		lds	ZH,libmio_ram+45
		ldi	XL,32			;bytes/line
		mul	XL,ctrl
		add	ZL,r0
		adc	ZH,r1
		adiw	ZL,32			;skip name
mem_readline0:	ldi	YL,LOW(bas_linebuf)	;buffer address
		ldi	YH,HIGH(bas_linebuf)
		ldi	XL,32			;no of bytes to copy
mem_readline1:	lpm	XH,Z+			;get from flash
		st	Y+,XH			;put to linebuffer
		dec	XL
		brne	mem_readline1
		ldi	XH,0xff			;append an extra EOL
		st	Y+,XH
		ret					

		
mem_spm:	in	tempreg1,SPMCSR
		sbrc	tempreg1,0
		rjmp	mem_spm
mem_spm1:	out	SPMCSR,XH
		spm
mem_spm2:	in	tempreg1,SPMCSR
		sbrc	tempreg1,0
		rjmp	mem_spm2
		ret
					
mem_waitrww:	in	tempreg1,SPMCSR
		sbrs	tempreg1,6
		ret
		ldi	XH,0x11
		rcall	mem_spm
		rjmp	mem_waitrww			


;------------------------------------------------------------------------------
; write line from Y to program memory
;------------------------------------------------------------------------------
mem_wrline:	cli
		mov	tempreg1,ZL		;pointer
		andi	tempreg1,0xe0		;mask out low bits
		brne	mem_wrline_1		;no first line
;erase page		
		ldi	XH,0x03			;page erase
		rcall	mem_spm
		rcall	mem_waitrww

mem_wrline_1:	ldi	XL,16			;16 words to copy
mem_wrline_2:	ld	r0,Y+			;low byte
		ld	r1,Y+			;high byte
		ldi	XH,0x01			;write to buffer
		rcall	mem_spm
		adiw	ZL,2			;ptr+2
		dec	XL			;loop counter
		brne	mem_wrline_2		;loop
;check for end of page						
		cpi	ZL,0x00			;end of page
		breq	mem_wrline_3		;yes
		ret
		
;write page
mem_wrline_3:	sbiw	ZL,2
		ldi	XH,0x05			;page write
		rcall	mem_spm
		rcall	mem_waitrww
		adiw	ZL,2
		ret
				

;------------------------------------------------------------------------------
; write array to program memory
;------------------------------------------------------------------------------
mem_writepage:	cli
		push	tempreg1
		push	XH
		push	XL
		ldi	XH,0x03			;page erase
		rcall	mem_spm
		rcall	mem_waitrww

		sbiw	ZL,2
		ldi	YL,LOW(bas_array+512)	;buffer address
		ldi	YH,HIGH(bas_array+512)
		ldi	XL,0x80			;no of words to copy
mem_wpage1:	adiw	ZL,2
		ld	r0,Y+			;low byte
		ld	r1,Y+			;high byte
		ldi	XH,0x01			;write to buffer
		rcall	mem_spm
		dec	XL
		brne	mem_wpage1
	
		ldi	XH,0x05			;page write
		rcall	mem_spm
		rcall	mem_waitrww
		pop	XL
		pop	XH
		pop	tempreg1
		ret


;------------------------------------------------------------------------------
; read name from program memory to Y+
;------------------------------------------------------------------------------
mem_readname:	lds	ZL,libmio_ram+44	;base address of program
		lds	ZH,libmio_ram+45
		sbiw	ZL,32			;point to fname
mem_readname0:	ldi	XL,12			;bytes/name
mem_readname1:	lpm	XH,Z+		
		st	Y+,XH			;get bytes
		dec	XL
		brne	mem_readname1
		lds	XL,libmio_config
		sbrc	XL,7
		ret
		ldi	XL,LOW(libmio_cols*libmio_rows-12)	;offset
		ldi	XH,HIGH(libmio_cols*libmio_rows-12)
		add	YL,XL
		adc	YH,XH
		ldi	XL,12
		lds	XH,libmio_color
mem_readname2:	st	Y+,XH
		dec	XL
		brne	mem_readname2		
		ret					

;------------------------------------------------------------------------------
; read name from program ctrl to Y+
;------------------------------------------------------------------------------
mem_readnamec:	ldi	ZL,LOW(bas_programs*2)	;base address
		ldi	ZH,HIGH(bas_programs*2)
		ldi	XL,LOW(3072)		;bytes/program
		ldi	XH,HIGH(3072)
		mov	tempreg1,ctrl
mem_readnamec1:	cpi	tempreg1,0
		breq	mem_readname0
		add	ZL,XL
		adc	ZH,XH
		dec	tempreg1
		rjmp	mem_readnamec1			


;------------------------------------------------------------------------------
; save program from editor to flash
;------------------------------------------------------------------------------
mem_save:	cli				;disable INT
		rcall	mem_progadr
		ldi	YL,LOW(bas_linebuf)	;base address
		ldi	YH,HIGH(bas_linebuf)
		ldi	XL,LOW(libmio_vram+11)
		ldi	XH,HIGH(libmio_vram+11)
		ldi	tempreg2,12		;bytes/name
mem_save_01:	ld	tempreg1,X+		;get from filename
		st	Y+,tempreg1		;write to linebuf
		dec	tempreg2
		brne	mem_save_01
		ldi	tempreg2,20		;add free space
		ldi	tempreg1,0x20		;space char
mem_save_02:	st	Y+,XH			;write to linebuf
		dec	tempreg2
		brne	mem_save_02		
		sbiw	YL,32
		rcall	mem_wrline		;write first line
		clr	ctrl			;line number
;save lines loop
mem_save_03:	ldi	XH,libmio_ecols		;bytes/vram-line
		mul	XH,ctrl
	    	ldi	XL,LOW(libmio_vram+3*libmio_ecols+3)	;lines start
		ldi	XH,HIGH(libmio_vram+3*libmio_ecols+3)
		add	XL,r0			;calc line address in vram
		adc	XH,r1
;copy line to bas_inbuf
		ldi	YL,LOW(bas_inbuf)	;text buffer address
		ldi	YH,HIGH(bas_inbuf)
		ldi	tempreg2,libmio_ecols-3	;number of chars to copy
mem_save_04:	ld	tempreg1,X+		;get from vram
		st	Y+,tempreg1
		dec	tempreg2
		brne	mem_save_04		
		st	Y+,byte0		;add an extra EOL
;shrink line
		ldi	tempreg2,libmio_ecols-3	;max chars
mem_save_05:	ld	tempreg1,-Y		;		
		cpi	tempreg1,0x20		;space
		brne	mem_save_06
		st	Y,byte0			;write EOL	
		dec	tempreg2
		brne	mem_save_05
mem_save_06:	call	clin_main		;compress line
		ldi	YL,LOW(bas_linebuf)	;buffer address
		ldi	YH,HIGH(bas_linebuf)
		rcall	mem_wrline		;write compressed line
;next line
		inc	ctrl			;line +1
		cpi	ctrl,bas_lines		;last
		brne	mem_save_03		;no
		sei				;enable interrupts
		ret    				;OK, end here

;------------------------------------------------------------------------------
; load program from flash to editor
;------------------------------------------------------------------------------
mem_load:	rcall	mem_progadr		;set Z to address
;read name
		ldi	XL,LOW(libmio_vram+11)
		ldi	XH,HIGH(libmio_vram+11)
		ldi	tempreg2,12		;bytes/name
mem_load_01:	lpm	tempreg1,Z+		;get from filename
		st	X+,tempreg1		;write to linebuf
		dec	tempreg2
		brne	mem_load_01
		adiw	ZL,20			;set ptr to first line
		clr	ctrl			;line number

;load lines loop
mem_load_02:	ldi	tempreg2,32		;bytes per line
		ldi	XL,LOW(bas_linebuf)
		ldi	XH,HIGH(bas_linebuf)
mem_load_03:	lpm	tempreg1,Z+		;get from flash
		st	X+,tempreg1		;write to linebuf
		dec	tempreg2
		brne	mem_load_03
;clear text buffer
		ldi	XL,LOW(bas_inbuf)	;text buffer address
		ldi	XH,HIGH(bas_inbuf)
		ldi	tempreg1,0x20		;space
		ldi	tempreg2,libmio_ecols
lce_load_04:	st	X+,tempreg1
		dec	tempreg2
		brne	lce_load_04		;loop
;expand line			
		push	ZH			;save registers
		push	ZL
		push	ctrl
		call 	xlin_main		;expand line
		pop	ctrl			;restore registers
		pop	ZL
		pop	ZH
;copy line to vram
		ldi	YL,LOW(bas_inbuf)	;text buffer address
		ldi	YH,HIGH(bas_inbuf)
		ldi	XH,libmio_ecols		;bytes/vram-line
		mul	XH,ctrl
	    	ldi	XL,LOW(libmio_vram+3*libmio_ecols)	;lines start
		ldi	XH,HIGH(libmio_vram+3*libmio_ecols)
		add	XL,r0			;calc line address
		adc	XH,r1
		ldi	tempreg2,libmio_ecols
lce_load_05:	ld	tempreg1,Y+		;get from inbuf
		st	X+,tempreg1		;store to vram
		dec	tempreg2
		brne	lce_load_05
		
;next line
		inc	ctrl			;line +1
		cpi	ctrl,bas_lines
		brne	mem_load_02				
		ret

;------------------------------------------------------------------------------
; calculate memory program address
;------------------------------------------------------------------------------
mem_progadr:	lds	tempreg1,libmio_lastmenu	;program to save
mem_progadr_0:	ldi	ZL,LOW(bas_programs*2)		;base
		ldi	ZH,HIGH(bas_programs*2)
		andi	tempreg1,0x07		;mask bits
		swap	tempreg1		;*16
		ldi	XL,0xc0			;*192
		mul	XL,tempreg1
		add	ZL,r0			;set ptr to start
		adc	ZH,r1			
		ret	


		